# Internal Library
import aepp
from aepp import connector
from copy import deepcopy
import pandas as pd
import logging


class Profile:
    """
    A class containing the different methods exposed on Customer Profile API.
    The API documentation is available:
    https://www.adobe.io/apis/experienceplatform/home/api-reference.html#!acpdr/swagger-specs/real-time-customer-profile.yaml
    """

    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(
        self,
        config: dict = aepp.config.config_object,
        header=aepp.config.header,
        loggingObject: dict = None,
        **kwargs,
    ):
        """
            This will instantiate the profile class
            Arguments:
                config : OPTIONAL : config object in the config module.
                header : OPTIONAL : header object  in the config module.
                loggingObject : OPTIONAL : logging object to log messages.
        kwargs:
            kwargsvaluewillupdatetheheader
        """
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(
            config_object=config,
            header=header,
            loggingEnabled=self.loggingEnabled,
            logger=self.logger,
        )
        self.header = self.connector.header
        self.header["Accept"] = "application/vnd.adobe.xdm+json"
        self.header.update(**kwargs)
        self.sandbox = self.connector.config["sandbox"]
        # same endpoint than segmentation
        self.endpoint = (
            aepp.config.endpoints["global"] + aepp.config.endpoints["segmentation"]
        )

    def getEntity(
        self,
        schema_name: str = "_xdm.context.profile",
        entityId: str = None,
        entityIdNS: str = None,
        mergePoliciyId: str = None,
        **kwargs,
    ) -> dict:
        """
        Returns an entity by ID or Namespace.
        Arguments:
            schema_name : REQUIRED : class name of the schema to be retrieved. default : _xdm.context.profile
            entityId : OPTIONAL : identity ID
            entityIdNS : OPTIONAL : Identity Namespace code. Required if entityId is used (except for native identity)
            mergePoliciyId : OPTIONAL : Id of the merge policy.
        Possible kwargs:
            fields : path of the elements to be retrieved, separated by comma. Ex : "person.name.firstName,person.name.lastName"
            relatedSchema_name : If schema.name is "_xdm.context.experienceevent", this value must specify the schema for the profile entity that the time series events are related to.
            relatedEntityId : ID of the entity that the ExperienceEvents are associated with. Used when looking up ExperienceEvents. For Native XID lookup, use relatedEntityId=<XID> and leave relatedEntityIdNS absent;
            For ID:NS lookup, use both relatedEntityId and relatedEntityIdNS fields.
            relatedEntityIdNS : Identity Namespace code of the related entity ID of ExperienceEvent. Used when looking up ExperienceEvents. If this field is used, entityId cannot be empty.
            startTime : Start time of Time range filter for ExperienceEvents. Should be at millisecond granularity. Included. Default: From beginning.
            endTime : End time of Time range filter for ExperienceEvents. Should be at millisecond granularity. Excluded. Default: To the end.
            limit : Number of records to return from the result. Only for time-series objects. Default: 1000
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getEntity")
        path = "/access/entities"
        params = {
            "schema.name": schema_name,
            "entityId": entityId,
            "entityIdNS": entityIdNS,
            "mergePoliciyId": mergePoliciyId,
        }
        if schema_name == "_xdm.context.experienceevent":
            params["relatedSchema.name"] = kwargs.get(
                "relatedSchema_name", "_xdm.context.profile"
            )
            params["relatedEntityId"] = kwargs.get("relatedEntityId", entityId)
            params["relatedEntityIdNS"] = kwargs.get("relatedEntityIdNS", None)
            params["limit"] = kwargs.get("limit", 1000)
            params["startTime"] = kwargs.get("startTime", None)
            params["endTime"] = kwargs.get("endTime", None)
        params["fields"] = kwargs.get("fields", None)
        res = self.connector.getData(
            self.endpoint + path, params=params, headers=self.header
        )
        return res

    def getEntities(self, request_data: dict = None) -> dict:
        """
        Get a number of different identities from ID or namespaces.
        Argument:
            request_data : Required : A dictionary that contains fields for the search.
            Example
            {
                "schema": {
                    "name": "_xdm.context.profile"
                },
                "relatedSchema": {
                    "name": "_xdm.context.profile"
                },
                "fields": [
                    "person.name.firstName"
                ],
                "identities": [
                    {
                    "entityId": "69935279872410346619186588147492736556",
                    "entityIdNS": {
                        "code": "CRMID"
                    }
                    }
                ],
                "timeFilter": {
                    "startTime": 1539838505,
                    "endTime": 1539838510
                },
                "limit": 10,
                "orderby": "-timestamp",
                "withCA": True
            }
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getEntities")
        path = "/access/entities"
        if request_data is None or type(request_data) != dict:
            raise Exception("Expected a dictionary to fetch entities")
        res = self.connector.postData(
            self.endpoint + path, data=request_data, headers=self.header
        )
        return res

    def deleteEntity(
        self, schema_name: str = None, entityId: str = None, entityIdNS: str = None
    ) -> str:
        """
        Delete a specific entity
        Arguments:
            schema_name : REQUIRED : Name of the associated XDM schema.
            entityId : REQUIRED : entity ID
            entityIdNS : OPTIONAL : entity ID Namespace
        """
        path = "/access/entities"
        params = {}
        if schema_name is None:
            raise Exception("Expected a schema name")
        else:
            params["schema.name"] = schema_name
        if entityId is not None:
            params["entityId"] = entityId
        if entityIdNS is not None:
            params["entityIdNS"] = entityIdNS
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteEntity")
        res = self.connector.deleteData(
            self.endpoint + path, params=params, headers=self.header
        )

        return res

    def getMergePolicies(self, limit: int = 100) -> dict:
        """
        Returns the list of merge policies hosted in this instance.
        Arguments:
            limit : OPTIONAL : amount of merge policies returned by pages.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getMergePolicies")
        path = "/config/mergePolicies"
        params = {"limit": limit}
        res = self.connector.getData(
            self.endpoint + path, params=params, headers=self.header
        )
        data = res["children"]
        nextPage = res["_links"]["next"].get("href", "")
        while nextPage != "":
            path = "/config/mergePolicies?" + nextPage.split("?")[1]
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=self.header
            )
            data += res["children"]
            nextPage = res["_links"]["next"].get("href", "")
        return data

    def getMergePolicy(self, policy_id: str = None) -> dict:
        """
        Return a specific merge policy.
        Arguments:
            policy_id : REQUIRED : id of the the policy id to be returned.
        """
        if policy_id is None:
            raise Exception("Missing the policy id")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getMergePolicy")
        path = f"/config/mergePolicies/{policy_id}"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        return res

    def createMergePolicy(self, policy: dict = None) -> dict:
        """
        Arguments:
            policy: REQUIRED : The dictionary defining the policy
        Refer to the documentation : https://experienceleague.adobe.com/docs/experience-platform/profile/api/merge-policies.html
        Example of JSON:
        {
            "name": "real-time-customer-profile-default",
            "imsOrgId": "1BD6382559DF0C130A49422D@AdobeOrg",
            "schema": {
                "name": "_xdm.context.profile"
            },
            "default": False,
            "identityGraph": {
                "type": "pdg"
            },
            "attributeMerge": {
                "type": "timestampOrdered",
                "order": [
                "string"
                ]
            },
            "updateEpoch": 1234567890
        }
        """
        path = "/config/mergePolicies"
        if policy is None:
            raise ValueError("Require a dictionary")
        if self.loggingEnabled:
            self.logger.debug(f"Starting createMergePolicy")
        res = self.connector.postData(
            self.endpoint + path, data=policy, headers=self.header
        )
        return res

    def updateMergePolicy(self, mergePolicyId: str = None, policy: dict = None) -> dict:
        """
        Update a merge policy by replacing its definition. (PUT method)
        Arguments:
            mergePolicyId : REQUIRED : The merge Policy Id
            policy : REQUIRED : a dictionary giving the definition of the merge policy
            Refer to the documentation : https://experienceleague.adobe.com/docs/experience-platform/profile/api/merge-policies.html
        Example of JSON:
        {
            "name": "real-time-customer-profile-default",
            "imsOrgId": "1BD6382559DF0C130A49422D@AdobeOrg",
            "schema": {
                "name": "_xdm.context.profile"
            },
            "default": False,
            "identityGraph": {
                "type": "pdg"
            },
            "attributeMerge": {
                "type": "timestampOrdered",
                "order": [
                "string"
                ]
            },
            "updateEpoch": 1234567890
        }
        attributeMerge can be "dataSetPrecedence" or "timestampOrdered". Please provide a list of dataSetId on "order" when using the first option.
        """
        if mergePolicyId is None:
            raise ValueError("Require a mergePolicyId")
        if policy is None or type(policy) != dict:
            raise ValueError("Require a dictionary to update the merge policy")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateMergePolicy")
        path = f"/config/mergePolicies/{mergePolicyId}"
        res = self.connector.putData(
            self.endpoint + path, data=policy, headers=self.header
        )
        return res

    def patchMergePolicy(
        self, mergePolicyId: str = None, operations: list = None
    ) -> str:
        """
        Update a merge policy by replacing its definition.
        Arguments:
            mergePolicyId : REQUIRED : The merge Policy Id
            operations : REQUIRED : a list of operations to realize on the merge policy
            Refer to the documentation : https://experienceleague.adobe.com/docs/experience-platform/profile/api/merge-policies.html
        Example of operation:
        [
            {
                "op": "add",
                "path": "/identityGraph.type",
                "value": "pdg"
            }
        ]
        """
        if mergePolicyId is None:
            raise ValueError("Require a mergePolicyId")
        if operations is None or type(operations) != list:
            raise ValueError("Require a dictionary to update the merge policy")
        if self.loggingEnabled:
            self.logger.debug(f"Starting patchMergePolicy")
        path = f"/config/mergePolicies/{mergePolicyId}"
        res = self.connector.patchData(
            self.endpoint + path, data=operations, headers=self.header
        )
        return res

    def deleteMergePolicy(self, mergePolicyId: str = None) -> str:
        """
        Delete a merge policy by its ID.
        Arguments:
            mergePolicyId : REQUIRED : The merge Policy to be deleted
        """
        if mergePolicyId is None:
            raise ValueError("Require a mergePolicyId")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteMergePolicy")
        path = f"/config/mergePolicies/{mergePolicyId}"
        res = self.connector.deleteData(self.endpoint + path)
        return res

    def getPreviewStatus(self) -> dict:
        """
        View the details for the last successful sample job that was run for the IMS Organization.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getPreviewStatus")
        path = "/previewsamplestatus"
        privateHeader = deepcopy(self.header)
        privateHeader["Accept"] = "application/json"
        res = self.connector.getData(self.endpoint + path, headers=privateHeader)
        return res

    def getPreviewDataSet(self, date: str = None, output: str = "raw") -> dict:
        """
        View a report showing the distribution of profiles by dataset.
        Arguments:
            date : OPTIONAL : Format: YYYY-MM-DD.
                If multiple reports were run on the date, the most recent report for that date will be returned.
                If a report does not exist for the specified date, a 404 error will be returned.
                If no date is specified, the most recent report will be returned.
                Example: date=2024-12-31
            output : OPTIONAL : if you want to have a dataframe returns. Use "df", default "raw"
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getPreviewDataSet")
        path = "/previewsamplestatus/report/dataset"
        params = {}
        if date is not None:
            params["date"] = date
        privateHeader = deepcopy(self.header)
        privateHeader["Accept"] = "application/json"
        res = self.connector.getData(self.endpoint + path, headers=privateHeader)
        if output == "df":
            df = pd.DataFrame(res["data"])
            return df
        return res

    def getPreviewDataSetOverlap(self, date: str = None, output: str = "raw") -> dict:
        """
        Method to find the overlap of identities with datasets.
        Arguments:
            date : OPTIONAL : Format: YYYY-MM-DD.
                If multiple reports were run on the date, the most recent report for that date will be returned.
                If a report does not exist for the specified date, a 404 error will be returned.
                If no date is specified, the most recent report will be returned.
                Example: date=2024-12-31
            output : OPTIONAL : if you want to have a dataframe returns. Use "df", default "raw"
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getPreviewDataSetOverlap")
        path = "/previewsamplestatus/report/dataset/overlap"
        params = {}
        if date is not None:
            params["date"] = date
        privateHeader = deepcopy(self.header)
        privateHeader["Accept"] = "application/json"
        privateHeader["x-model-name"] = "_xdm.context.profile"
        res = self.connector.getData(self.endpoint + path, headers=privateHeader)
        if output == "df":
            df = pd.DataFrame(res["data"])
            return df
        return res

    def getPreviewNamespace(self, date: str = None, output: str = "raw") -> dict:
        """
        View a report showing the distribution of profiles by namespace.
        Arguments:
            date : OPTIONAL : Format: YYYY-MM-DD.
                If multiple reports were run on the date, the most recent report for that date will be returned.
                If a report does not exist for the specified date, a 404 error will be returned.
                If no date is specified, the most recent report will be returned.
                Example: date=2024-12-31
            output : OPTIONAL : if you want to have a dataframe returns. Use "df", default "raw"
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getPreviewNamespace")
        path = "/previewsamplestatus/report/namespace"
        params = {}
        if date is not None:
            params["date"] = date
        privateHeader = deepcopy(self.header)
        privateHeader["Accept"] = "application/json"
        res = self.connector.getData(self.endpoint + path, headers=privateHeader)
        if output == "df":
            df = pd.DataFrame(res["data"])
            return df
        return res

    def createDeleteSystemJob(self, dataSetId: str = None, batchId: str = None) -> dict:
        """
        Delete all the data for a batch or a dataSet based on their ids.
        Note: you cannot delete batch from record type dataset. You can overwrite them to correct the issue.
        Only Time Series and record type datasets can be deleted.
        Arguments:
            dataSetId : REQUIRED : dataSetId to be deleted
            batchId : REQUIRED : batchId to be deleted.
        More info: https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Profile_System_Jobs/createDeleteRequest
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting createDeleteSystemJob")
        path = "/system/jobs"
        if dataSetId is not None:
            obj = {"dataSetId": dataSetId}
            res = self.connector.postData(self.endpoint + path, data=obj)
            return res
        elif batchId is not None:
            obj = {"batchId": batchId}
            res = self.connector.postData(self.endpoint + path, data=obj)
            return res
        else:
            raise ValueError("Require a dataSetId or a batchId")

    def getDeleteSystemJobs(
        self, page: int = 0, limit: int = 100, n_results: int = 100
    ) -> dict:
        """
        Retrieve a list of all delete requests (Profile System Jobs) created by your organization.
        Arguments:
            page : OPTIONAL : Return a specific page of results, as per the create time of the request. For example, page=0
            limit : OPTIONAL : Limit response to a specific number of objects. Must be a positive number. For example, limit=10
            n_results : OPTIONAL : Number of total result to retrieve.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getDeleteSystemJobs")
        path = "/system/jobs"
        params = {"page": page, "limit": limit}
        res = self.connector.getData(self.endpoint + path, params=params)
        data = res["children"]
        count = len(data)
        nextPage = res["_page"].get("next", "")
        while nextPage != "" and count < n_results:
            page += 1
            params = {"page": page, "limit": limit}
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=self.header
            )
            count += len(res["children"])
            data += res["children"]
            nextPage = res["_page"].get("next", "")
        return data

    def getDeleteSystemJob(self, jobId: str = None) -> dict:
        """
        Get a specific delete system job by its ID.
        Arguments:
            jobId : REQUIRED : Job ID to be retrieved.
        """
        if jobId is None:
            raise ValueError("Require a system Job ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getDeleteSystemJob")
        path = f"/system/jobs/{jobId}"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        return res

    def deleteDeleteSystemJob(self, jobId: str = None) -> dict:
        """
        Delete a specific delete system job by its ID.
        Arguments:
            jobId : REQUIRED : Job ID to be deleted.
        """
        if jobId is None:
            raise ValueError("Require a system Job ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteDeleteSystemJob")
        path = f"/system/jobs/{jobId}"
        res = self.connector.deleteData(self.endpoint + path, headers=self.header)
        return res

    def getComputedAttributes(self) -> list:
        """
        Retrieve the list of computed attributes set up in your organization.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getComputedAttributes")
        path = "/config/computedAttributes"
        res = self.connector.getData(self.endpoint + path)
        data = res["children"]
        nextPage = res["_page"].get("next", "")
        # while nextPage != "":
        #     res = self.connector.getData(self.endpoint+path,
        #                     params=params, headers=self.header)
        #     data += res['children']
        #     nextPage = res['_page'].get('next','')
        return res

    def getComputedAttribute(self, attributeId: str = None) -> dict:
        """
        Returns a single computed attribute.
        Arguments:
            attributeId : REQUIRED : The computed attribute ID.
        """
        if attributeId is None:
            raise ValueError("Require a computed attribute ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getComputedAttribute")
        path = f"/config/computedAttributes/{attributeId}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def deleteComputedAttribute(self, attributeId: str = None) -> dict:
        """
        Delete a specific computed attribute.
        Arguments:
            attributeId : REQUIRED : The computed attribute ID to be deleted.
        """
        if attributeId is None:
            raise ValueError("Require a computed attribute ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteComputedAttributes")
        path = f"/config/computedAttributes/{attributeId}"
        res = self.connector.deleteData(self.endpoint + path)
        return res

    def getDestinations(self) -> dict:
        """
        Retrieve a list of edge projection destinations. The latest definitions are returned.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getDestinations")
        path = "/config/destinations"
        res = self.connector.getData(self.endpoint + path)
        return res

    def createDestination(self, destinationConfig: dict = None) -> dict:
        """
        Create a new edge projection destination. Assume that there is time between creation and propagation of this information to the edge.
        Arguments:
            destinationConfig : REQUIRED : the destination configuration
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting createDestination")
        privateHeader = deepcopy(self.header)
        privateHeader[
            "Content-Type"
        ] = "application/vnd.adobe.platform.projectionDestination+json; version=1"
        if destinationConfig is None:
            raise ValueError("Require a destination configuration object")
        path = "/config/destinations"
        res = self.connector.postData(
            self.endpoint + path, data=destinationConfig, headers=privateHeader
        )
        return res

    def getDestination(self, destinationId: str = None) -> dict:
        """
        Get a specific destination based on its ID.
        Arguments:
            destinationId : REQUIRED : The destination ID to be retrieved
        """
        if destinationId is None:
            raise ValueError("Require a destination ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getDestination")
        path = f"/config/destinations/{destinationId}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def deleteDestination(self, destinationId: str = None) -> dict:
        """
        Delete a specific destination based on its ID.
        Arguments:
            destinationId : REQUIRED : The destination ID to be deleted
        """
        if destinationId is None:
            raise ValueError("Require a destination ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteDestination")
        path = f"/config/destinations/{destinationId}"
        res = self.connector.deleteData(self.endpoint + path)
        return res

    def updateDestination(
        self, destinationId: str = None, destinationConfig: dict = None
    ) -> dict:
        """
        Update a specific destination based on its ID. (PUT request)
        Arguments:
            destinationId : REQUIRED : The destination ID to be updated
            destinationConfig : REQUIRED : the destination config object to replace the old one.
        """
        if destinationId is None:
            raise ValueError("Require a destination ID")
        if destinationConfig is None:
            raise ValueError("Require a dictionation for updating the destination")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateDestination")
        privateHeader = deepcopy(self.header)
        privateHeader[
            "Content-Type"
        ] = "application/vnd.adobe.platform.projectionDestination+json"
        path = f"/config/destinations/{destinationId}"
        res = self.connector.putData(
            self.endpoint + path, data=destinationConfig, headers=privateHeader
        )
        return res

    def getProjections(self, schemaName: str = None, name: str = None) -> dict:
        """
        Retrieve a list of edge projection configurations. The latest definitions are returned.
        Arguments:
            schemaName : OPTIONAL : The name of the schema class associated with the projection configuration you want to access.
                example : _xdm.context.profile
            name : OPTIONAL : The name of the projection configuration you want to access.
                if name is specified, schemaName is also required.
        """
        path = "/config/projections"
        params = {}
        if name is not None and schemaName is None:
            raise AttributeError(
                "You must specify a schema name when setting a projection name"
            )
        if self.loggingEnabled:
            self.logger.debug(f"Starting getProjections")
        if schemaName is not None:
            params["schemaName"] = schemaName
        if name is not None:
            params["name"] = name
        res = self.connector.getData(
            self.endpoint + path, params=params, headers=self.headers
        )
        return res

    def getProjection(self, projectionId: str = None) -> dict:
        """
        Retrieve a single projection based on its ID.
        Arguments:
            projectionId : REQUIRED : the projection ID to be retrieved.
        """
        if projectionId is None:
            raise ValueError("Require a projection ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getProjection")
        path = f"/config/projections/{projectionId}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def deleteProjection(self, projectionId: str = None) -> dict:
        """
        Delete a single projection based on its ID.
        Arguments:
            projectionId : REQUIRED : the projection ID to be deleted.
        """
        if projectionId is None:
            raise ValueError("Require a projection ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteProjection")
        path = f"/config/projections/{projectionId}"
        res = self.connector.deleteData(self.endpoint + path)
        return res

    def createProjection(
        self, schemaName: str = None, projectionConfig: dict = None
    ) -> dict:
        """
        Create a projection
        Arguments:
            schemaName : REQUIRED : XDM schema namess
            projectionConfig : REQUIRED : the object definiing the projection
        """
        if schemaName is None:
            raise ValueError("Require a schema name specified")
        if projectionConfig is None:
            raise ValueError("Require a projection configuration")
        if self.loggingEnabled:
            self.logger.debug(f"Starting createProjection")
        path = "/config/projections"
        params = {"schemaName": schemaName}
        privateHeader = deepcopy(self.header)
        privateHeader[
            "Content-Type"
        ] = "application/vnd.adobe.platform.projectionConfig+json; version=1"
        res = self.connector.postData(
            self.endpoint + path,
            params=params,
            data=projectionConfig,
            privateHeader=privateHeader,
        )
        return res

    def updateProjection(
        self, projectionId: str = None, projectionConfig: dict = None
    ) -> dict:
        """
        Update a projection based on its ID.(PUT request)
        Arguments:
            projectionId : REQUIRED : The ID of the projection to be updated.
            projectionConfig : REQUIRED : the object definiing the projection
        """
        if projectionId is None:
            raise ValueError("Require a projectionId")
        if projectionConfig is None:
            raise ValueError("Require a projection Configuration object")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateProjection")
        privateHeader = deepcopy(self.header)
        privateHeader[
            "Content-Type"
        ] = "application/vnd.adobe.platform.projectionConfig+json"
        path = f"/config/projections/{projectionId}"
        res = self.connector.putData(
            self.endpoint + path, data=projectionConfig, headers=privateHeader
        )
        return res



import aepp
from dataclasses import dataclass
from aepp import connector
import pandas as pd
from copy import deepcopy
from typing import Union
import time
import codecs
import json
import logging

@dataclass
class _Data:

    def __init__(self):
        self.table_names = {}
        self.schema_ref = {}
        self.ids = {}


class Catalog:
    """
    Catalog class from the AEP API. This class helps you to find where the data are coming from in AEP.
    More details here : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#
    It possess a data attribute that is containing information about your datasets. 
    Arguments:
        config : OPTIONAL : config object in the config module (DO NOT MODIFY)
        header : OPTIONAL : header object  in the config module (DO NOT MODIFY)
        loggingObject : OPTIONAL : If you want to set logging capability for your actions.
    kwargs:
        kwargs value will update the header
    """
    loggingEnabled = False
    logger = None

    def __init__(self,
                config:dict=aepp.config.config_object,
                header=aepp.config.header,
                loggingObject:dict=None,
                **kwargs):
        if loggingObject is not None and sorted(["level","stream","format","filename","file"]) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(config_object=config, header=header,loggingEnabled=self.loggingEnabled,logger=self.logger)
        self.header = self.connector.header
        self.header.update(**kwargs)
        self.sandbox = self.connector.config['sandbox']
        self.endpoint = aepp.config.endpoints['global']+aepp.config.endpoints["catalog"]
        self.data = _Data()

    def getResource(self,endpoint:str=None,params:dict=None,format:str='json',save:bool=False,**kwargs)->dict:
        """
        Template for requesting data with a GET method.
        Arguments:
            endpoint : REQUIRED : The URL to GET
            params: OPTIONAL : dictionary of the params to fetch
            format : OPTIONAL : Type of response returned. Possible values:
                json : default
                txt : text file
                raw : a response object from the requests module
        """
        if endpoint is None:
            raise ValueError("Require an endpoint")
        if self.loggingEnabled:
            self.logger.debug(f"Using getResource with following format ({format}) to the following endpoint: {endpoint}")
        res = self.connector.getData(endpoint,params=params,format=format)
        if save:
            if format == 'json':
                aepp.saveFile(module="catalog",file=res,filename=f"resource_{int(time.time())}",type_file="json",encoding=kwargs.get("encoding",'utf-8'))
            elif format == 'txt':
                aepp.saveFile(module="catalog",file=res,filename=f"resource_{int(time.time())}",type_file="txt",encoding=kwargs.get("encoding",'utf-8'))
            else:
                print("element is an object. Output is unclear. No save made.\nPlease save this element manually")
        return res
    
    def decodeStreamBatch(self,message:str)->dict:
        """
        Decode the full txt batch via the codecs module.
        Usually the full batch is returned by the getResource method with format == "txt".
        Arguments:
            message: REQUIRED : the text file return from the failed batch message.
        
        return None when issue is raised
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting decodeStreamBatch")
        try: 
            decodeMessage = codecs.escape_decode(message)[0].decode().replace('"body":"{','"body":{').replace('}","header":"{','},"header":{').replace('}","_errors":"{','},"_errors":{').replace('}"','}')
            return decodeMessage
        except:
            print("Issue decoding the message.")
            return None

    def jsonStreamMessages(self,message:str,verbose:bool = False)->list:
        """
        Try to create a list of dictionary messages from the decoded stream batch extracted from the decodeStreamBatch method.
        Arguments:
            message : REQUIRED : a decoded text file, usually returned from the decodeStreamBatch method
            verbose : OPTIONAL : print errors and information on the decoding.
        
        return None when issue is raised
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting jsonStreamMessages")
        try:
            myList = []
            myYield:iter = (line for line in message.split("\n"))
            countLine,countErrors = 0,0
            for element in myYield:
                countLine +=1
                try:
                    myList.append(json.loads(element))

                except Exception as e:
                    countErrors+=1
                    if verbose:
                        print(e)
            if verbose:
                print(f"error rate is {(countErrors/countLine)*100:.2f}%")
            return myList
        except:
            print("Issue creating a stream of messages.")
            if self.loggingEnabled:
                self.logger.info(f"Issue creating a stream of messages")
            return None

    def getBatches(self,limit:int=10, n_results:int=None,output:str='raw',**kwargs)->Union[pd.DataFrame,dict]:
        """
        Retrieve a list of batches.
        Arguments:
            limit : Limit response to a specified positive number of objects. Ex. limit=10 (max = 100)
            n_results : OPTIONAL :  number of result you want to get in total. (will loop)
            output : OPTIONAL : Can be "raw" response (dict) or "dataframe".
        Possible kwargs:
            created : Filter by the Unix timestamp (in milliseconds) when this object was persisted.
            createdAfter : Exclusively filter records created after this timestamp. 
            createdBefore : Exclusively filter records created before this timestamp.
            start : Returns results from a specific offset of objects. This was previously called offset. (see next line)
                offset : Will offset to the next limit (sort of pagination)        
            updated : Filter by the Unix timestamp (in milliseconds) for the time of last modification.
            createdUser : Filter by the ID of the user who created this object.
            dataSet : Used to filter on the related object: &dataSet=dataSetId.
            version : Filter by Semantic version of the account. Updated when the object is modified.
            status : Filter by the current (mutable) status of the batch.
            orderBy : Sort parameter and direction for sorting the response. 
                Ex. orderBy=asc:created,updated. This was previously called sort.
            property : A comma separated whitelist of top-level object properties to be returned in the response. 
                Used to cut down the number of properties and amount of data returned in the response bodies.
            size : The number of bytes processed in the batch.
        # /Batches/get_batch
        more details : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        path = "/batches"
        limit = min([limit,100])
        params = {'limit':limit,**kwargs}
        if self.loggingEnabled:
            self.logger.debug(f"Starting getBatches with output format {output}")
        ## looping to retrieve pagination
        if n_results is not None:
            list_return = {}
            params['start'] = 0
            res = self.connector.getData(self.endpoint+path,
                            headers=self.header, params=params)
            list_return.update(**res)
            while len(list_return) < n_results and len(res) != 0:
                params['start'] += limit
                res = self.connector.getData(self.endpoint+path,
                            headers=self.header, params=params)
                list_return.update(**res)
            if output=="dataframe":
                return pd.DataFrame(list_return).T
            return list_return
        res = self.connector.getData(self.endpoint+path,
                            headers=self.header, params=params)
        if output=="dataframe":
            return pd.DataFrame(res).T
        return res

    def getFailedBatchesDF(self,limit:int=10,n_results: str=None)->pd.DataFrame:
        """
        Abstraction of getBatches method that focus on failed batches and return a dataframe with the batchId and errors.
        Also adding some meta data information from the batch information provided.
        Arguments:
            limit : Limit response to a specified positive number of objects. Ex. limit=10 (max = 100)
            n_results : OPTIONAL :  number of result you want to get in total. (will loop)
        """
        res = self.getBatches(status="failed",orderBy="desc:created",limit=limit,n_results=n_results)
        if self.loggingEnabled:
            self.logger.debug(f"Starting getFailedBatchesDF")
        dict_failed = {}
        for batch in res:
            if res[batch]['relatedObjects'][0]['type'] == "dataSet":
                datasetId = res[batch]['relatedObjects'][0]['id']
            dict_failed[batch] = {
                "timestamp" : res[batch]['created'],
                "recordsSize" : res[batch].get('metrics',{}).get('recordsSize',0),
                "invalidRecordsProfile" : res[batch].get('metrics',{}).get('invalidRecordsProfile',0),
                "invalidRecordsIdentity" : res[batch].get('metrics',{}).get('invalidRecordsIdentity',0),
                "invalidRecordCount" : res[batch].get('metrics',{}).get('invalidRecordCount',0),
                "invalidRecordsStreamingValidation" : res[batch].get('metrics',{}).get('invalidRecordsStreamingValidation',0),
                "invalidRecordsMapper" : res[batch].get('metrics',{}).get('invalidRecordsMapper',0),
                "invalidRecordsUnknown" : res[batch].get('metrics',{}).get('invalidRecordsUnknown',0),
                "errorCode" : res[batch]['errors'][0]['code'],
                "errorMessage" : res[batch]['errors'][0]['description'] ,
                "flowId" : res[batch].get('tags',{}).get('flowId',''),
                "dataSetId" : datasetId,
                "sandbox" : res[batch]['sandboxId'],
            }
        df = pd.DataFrame(dict_failed).T
        return df

    def getBatch(self, batch_id: str = None)->dict:
        """
        Get a specific batch id.
        Arguments:
            batch_id : REQUIRED : batch ID to be retrieved.
        """
        if batch_id is None:
            raise Exception("batch_id parameter is required.")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getBatch for the following batch : {batch_id}")
        path = f"/batches/{batch_id}"
        res = self.connector.getData(self.endpoint+path,
                            headers=self.header)
        return res
    
    def createBatch(self, object:dict=None,**kwargs) -> dict:
        """
        Create a new batch.
        Arguments:
            object : REQUIRED : Object that define the data to be onboarded.
                see reference here: https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Batches/postBatch
        """
        if object is None:
            raise Exception('expecting a definition of the data to be uploaded.')
        if self.loggingEnabled:
            self.logger.debug(f"Starting createBatch")
        path = "/batches"
        res = self.connector.postData(self.endpoint+path,data=object,
                            headers=self.header)
        return res

    def getResources(self, **kwargs)->list:
        """
        Retrieve a list of resource links for the Catalog Service.
        Possible kwargs:
            limit : Limit response to a specified positive number of objects. Ex. limit=10
            orderBy : Sort parameter and direction for sorting the response. 
                Ex. orderBy=asc:created,updated. This was previously called sort.
            property : A comma separated whitelist of top-level object properties to be returned in the response. 
                Used to cut down the number of properties and amount of data returned in the response bodies.
        """
        path = "/"
        params = {**kwargs}
        if self.loggingEnabled:
            self.logger.debug(f"Starting getResources")
        res = self.connector.getData(self.endpoint+path,
                            headers=self.header, params=params)
        return res


    def getDataSets(self,limit:int=100,output:str="raw",**kwargs)->dict:
        """
        Return a list of a datasets.
        Arguments:
            limit : REQUIRED : amount of dataset to be retrieved per call. 
            output : OPTIONAL : Default is "raw", other option is "df" for dataframe output
        Possible kwargs:
            state : The state related to a dataset.
            created : Filter by the Unix timestamp (in milliseconds) when this object was persisted.
            updated : Filter by the Unix timestamp (in milliseconds) for the time of last modification.
            name : Filter by the a descriptive, human-readable name for this DataSet.
            namespace : One of the registered platform acronyms that identify the platform.
            version : Filter by Semantic version of the account. Updated when the object is modified.
            property : Regex used to filter objects in the response. Ex. property=name~^test.
            # /Datasets/get_data_sets
            more possibilities : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
        """
        path = "/dataSets"
        params = {"limit":limit,**kwargs}
        if self.loggingEnabled:
            self.logger.debug(f"Starting getDataSets")
        res = self.connector.getData(self.endpoint+path, params=params)
        data = deepcopy(res)
        ## prepare pagination if needed
        start = 1
        while len(res) == limit:
            start +=limit
            params = {"limit":limit,"start":start,**kwargs}
            res = self.connector.getData(self.endpoint+path, params=params)
            data.update(res)
        try:
            if self.loggingEnabled:
                self.logger.debug(f"Starting caching results")
            self.data.table_names = {
                data[key]['name']: data[key]['tags']['adobe/pqs/table'] for key in data}
            self.data.schema_ref = {
                data[key]['name']: data[key]['schemaRef']
                for key in data if 'schemaRef' in data[key].keys()
            }
            self.data.ids = {
                data[key]['name']: key for key in data
            }
        except Exception as e:
            if self.loggingEnabled:
                self.logger.warning(f"Error caching results : {e}")
            print(e)
            print("Couldn't populate the data object from the instance.")
        if output == "df":
            df = pd.DataFrame(data).T
            return df
        return data

    def createDataSets(self, name:str=None, schemaId:str=None, data: dict = None, **kwargs):
        """
        Create a new dataSets based either on preconfigured setup or by passing the full dictionary for creation.
        Arguments:
            name : REQUIRED : if you wish to create a dataset via autocompletion. Provide a name.
            schemaId : REQUIRED : The schema $id reference for creating your dataSet.
            data : REQUIRED : If you want to pass the dataset object directly (not require the name and schemaId then)
                more info: https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Datasets/postDataset
        possible kwargs
            requestDataSource : Set to true if you want Catalog to create a dataSource on your behalf; otherwise, pass a dataSourceId in the body.
        """
        path = "/dataSets"
        params = {"requestDataSource": kwargs.get("requestDataSource", False)}
        if self.loggingEnabled:
            self.logger.debug(f"Starting createDataSets")
        if data is not None or isinstance(data, dict) == True:
            res = self.connector.postData(self.endpoint+path, params=params,
                             data=data)
        elif name is not None and schemaId is not None:
            data = {
                "name":name,
                "schemaRef": {
                    "id": schemaId,
                    "contentType": "application/vnd.adobe.xed+json;version=1"
                },
                "fileDescription": {
                    "persisted": True,
                    "containerFormat": "parquet",
                    "format": "parquet"
                }
            }
            res = self.connector.postData(self.endpoint+path, params=params,
                             data=data)
        return res

    def getDataSet(self, datasetId: str = None):
        """
        Return a single dataset.
        Arguments:
            datasetId : REQUIRED : Id of the dataset to be retrieved.
        """
        if datasetId is None:
            raise Exception("Expected a datasetId argument")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getDataset for : {datasetId}")
        path = f"/dataSets/{datasetId}"
        res = self.connector.getData(self.endpoint+path, headers=self.header)
        return res

    def deleteDataSet(self, datasetId: str = None):
        """
        Delete a dataset by its id.
        Arguments:
            datasetId : REQUIRED : Id of the dataset to be deleted.
        """
        if datasetId is None:
            raise Exception("Expected a datasetId argument")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteDataset for : {datasetId}")
        path = f"/dataSets/{datasetId}"
        res = self.connector.deleteData(self.endpoint+path, headers=self.header)
        return res

    ## Apparently deprecated.
    def getDataSetViews(self, datasetId: str = None, **kwargs):
        """
        Get views of the datasets.
        Arguments:
            datasetId : REQUIRED : Id of the dataset to be looked down.
        Possible kwargs:
            limit : Limit response to a specified positive number of objects. Ex. limit=10
            orderBy : Sort parameter and direction for sorting the response. Ex. orderBy=asc:created,updated.
            start : Returns results from a specific offset of objects. This was previously called offset. Ex. start=3.
            property : Regex used to filter objects in the response. Ex. property=name~^test.
        """
        if datasetId is None:
            raise Exception("Expected a datasetId argument")
        path = f"/dataSets/{datasetId}/views"
        params = {**kwargs}
        res = self.connector.getData(self.endpoint+path, headers=self.header)
        return res

    def getDataSetView(self, datasetId: str = None, viewId: str = None):
        """
        Get a specific view on a specific dataset.
        Arguments:
            datasetId : REQUIRED : ID of the dataset to be looked down.
            viewId : REQUIRED : ID of the view to be look upon.
        """
        if datasetId is None or viewId is None:
            raise Exception("Expected a datasetId and an viewId argument")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getDataSetView for datasetId: {datasetId} & viewId: {viewId}")
        path = f"/dataSets/{datasetId}/views/{viewId}"
        res = self.connector.getData(self.endpoint+path, headers=self.header)
        return res

    def getDataSetViewFiles(self, datasetId: str = None, viewId: str = None):
        """
        Returns the list of files attached to a view in a Dataset.
        Arguments:
            datasetId : REQUIRED : ID of the dataset to be looked down.
            viewId : REQUIRED : ID of the view to be look upon.
        """
        if datasetId is None or viewId is None:
            raise Exception("Expected a datasetId and an viewId argument")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getDataSetViewFiles for datasetId: {datasetId} & viewId: {viewId}")
        path = f"/dataSets/{datasetId}/views/{viewId}/files"
        res = self.connector.getData(self.endpoint+path, headers=self.header)
        return res



import aepp
from aepp import connector
import time
from concurrent import futures
import pandas as pd
import logging


class Observability:
    """
    A class that presents the different methods available on the Observability API from AEP.
    A complete documentation of the methods can be found here:
    https://www.adobe.io/apis/experienceplatform/home/api-reference.html#!acpdr/swagger-specs/observability-insights.yaml
    """

    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(
        self,
        config: dict = aepp.config.config_object,
        header=aepp.config.header,
        loggingObject: dict = None,
        **kwargs,
    ) -> None:
        """
        Instanciate the observability API methods class.
        Arguments:
            loggingObject : OPTIONAL : logging object to log messages.
            config : OPTIONAL : config object in the config module. (DO NOT MODIFY)
            header : OPTIONAL : header object  in the config module. (DO NOT MODIFY)
        """
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(
            config_object=config,
            header=header,
            loggingEnabled=self.loggingEnabled,
            logger=self.logger,
        )
        self.header = self.connector.header
        self.header.update(**kwargs)
        self.sandbox = self.connector.config["sandbox"]
        self.endpoint = (
            aepp.config.endpoints["global"] + aepp.config.endpoints["observability"]
        )
        self.POST_METRICS = {
            "start": "2020-07-14T00:00:00.000Z",
            "end": "2020-07-22T00:00:00.000Z",
            "granularity": "day",
            "metrics": [
                {
                    "name": "timeseries.ingestion.dataset.recordsuccess.count",
                    "filters": [
                        {
                            "name": "dataSetId",
                            "value": "5edcfb2fbb642119194c7d94|5eddb21420f516191b7a8dad",
                            "groupBy": True,
                        }
                    ],
                    "aggregator": "sum",
                    "downsample": "sum",
                },
                {
                    "name": "timeseries.ingestion.dataset.dailysize",
                    "filters": [
                        {
                            "name": "dataSetId",
                            "value": "5eddb21420f516191b7a8dad",
                            "groupBy": False,
                        }
                    ],
                    "aggregator": "sum",
                    "downsample": "sum",
                },
            ],
        }
        self._loadREFERENCES()

    def _loadREFERENCES(self):
        """
        Load document as attributes if possible
        """
        if self.loggingEnabled:
            self.logger.debug(f"Loading references")
        try:
            import importlib.resources as pkg_resources

            pathIdentity = pkg_resources.path("aepp", "observability_identity.pickle")
            pathIngestion = pkg_resources.path("aepp", "observability_ingestion.pickle")
            pathGDPR = pkg_resources.path("aepp", "observability_gdpr.pickle")
            pathQS = pkg_resources.path("aepp", "observability_queryService.pickle")
            pathRLTime = pkg_resources.path("aepp", "observability_realTime.pickle")
        except ImportError:
            # Try backported to PY<37 with pkg_resources.
            if self.loggingEnabled:
                self.logger.debug(f"Loading references - ImportError - 2nd try")
            try:
                import pkg_resources

                pathIdentity = pkg_resources.resource_filename(
                    "aepp", "observability_identity.pickle"
                )
                pathIngestion = pkg_resources.resource_filename(
                    "aepp", "observability_ingestion.pickle"
                )
                pathGDPR = pkg_resources.resource_filename(
                    "aepp", "observability_gdpr.pickle"
                )
                pathQS = pkg_resources.resource_filename(
                    "aepp", "observability_queryService.pickle"
                )
                pathRLTime = pkg_resources.resource_filename(
                    "aepp", "observability_realTime.pickle"
                )
            except:
                print("no supported files")
                if self.loggingEnabled:
                    self.logger.debug(f"Failed loading references")
        try:
            with pathIdentity as f:
                self.REFERENCE_IDENTITY = pd.read_pickle(f)
                self.REFERENCE_IDENTITY = self.REFERENCE_IDENTITY.style.set_properties(
                    subset=["Insights metric"], **{"width": "100px"}
                )
            with pathIngestion as f:
                self.REFERENCE_INGESTION = pd.read_pickle(f)
                self.REFERENCE_INGESTION = (
                    self.REFERENCE_INGESTION.style.set_properties(
                        subset=["Insights metric"], **{"width": "100px"}
                    )
                )
            with pathGDPR as f:
                self.REFERENCE_GDPR = pd.read_pickle(f)
                self.REFERENCE_GDPR = self.REFERENCE_GDPR.style.set_properties(
                    subset=["Insights metric"], **{"width": "100px"}
                )
            with pathRLTime as f:
                self.REFERENCE_REALTIME = pd.read_pickle(f)
                self.REFERENCE_REALTIME = self.REFERENCE_REALTIME.style.set_properties(
                    subset=["Insights metric"], **{"width": "100px"}
                )
            with pathQS as f:
                self.REFERENCE_QUERYSERVICE = pd.read_pickle(f)
                self.REFERENCE_QUERYSERVICE = (
                    self.REFERENCE_QUERYSERVICE.style.set_properties(
                        subset=["Insights metric"], **{"width": "100px"}
                    )
                )
        except:
            if self.loggingEnabled:
                self.logger.debug(f"Failed loading references - backup to None")
            self.REFERENCE_IDENTITY = None
            self.REFERENCE_INGESTION = None
            self.REFERENCE_QUERYSERVICE = None
            self.REFERENCE_GDPR = None
            self.REFERENCE_REALTIME = None

    def createMetricsReport(self, data: dict = None) -> dict:
        """
        Using the POST method to retrieve metrics specified in the data dictionary.
        Please use the different REFERENCES attributes to know which metrics are supported.
        You have a template for the data dictionary on the POST_METRICS attribute.
        Arguments:
            data : REQUIRED : The metrics requested in the report creation.
                You can use the POST_METRICS attribute to see a template.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting createMetricsReport")
        path = "/metrics"
        res = self.connector.postData(self.endpoint + path, data=data)
        return res



import aepp
import typing
from aepp import connector
import logging


class Policy:
    """
    Class to manage and retrieve DULE policy.
    This is based on the following API reference : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/
    """

    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(
        self,
        config: dict = aepp.config.config_object,
        header: dict = aepp.config.header,
        loggingObject: dict = None,
        **kwargs,
    ):
        """
        Instantiate the class to manage DULE rules and statement directly.
        Arguments:
            config : OPTIONAL : config object in the config module. (DO NOT MODIFY)
            header : OPTIONAL : header object  in the config module. (DO NOT MODIFY)
            loggingObject : OPTIONAL : logging object to log messages.
        """
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(
            config_object=config,
            header=header,
            loggingEnabled=self.loggingEnabled,
            loggingObject=self.logger,
        )
        self.header = self.connector.header
        self.header.update(**kwargs)
        self.sandbox = self.connector.config["sandbox"]
        self.endpoint = (
            aepp.config.endpoints["global"] + aepp.config.endpoints["policy"]
        )

    def getEnabledCorePolicies(self) -> dict:
        """
        Retrieve a list of all enabled core policies.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getEnabledCorePolicies")
        path = "/enabledCorePolicies"
        res = self.connector.getData(self.endpoint + path)
        return res

    def createEnabledCorePolicies(self, policyIds: list) -> dict:
        """
        Create or update the list of enabled core policies. (PUT)
        Argument:
            policyIds : REQUIRED : list of core policy ID to enable
        """
        path = "/enabledCorePolicies"
        if policyIds is None or type(policyIds) != list:
            raise ValueError("Require a list of policy ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting createEnabledCorePolicies")
        obj = policyIds
        res = self.connector.putData(self.endpoint + path, data=obj)
        return res

    def bulkEval(self, data: dict = None) -> dict:
        """
        Enable to pass a list of policies to check against a list of dataSet.
        Argument:
            data : REQUIRED : Dictionary describing the set of label and datasets.
                see https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Bulk_evaluation/bulkEvalPost
        """
        path = "/bulk-eval"
        if data is None:
            raise Exception(
                "Requires a dictionary to set the labels and dataSets to check"
            )
        if self.loggingEnabled:
            self.logger.debug(f"Starting bulkEval")
        res = self.connector.postData(self.endpoint + path, data=data)
        return res

    def getPoliciesCore(self, **kwargs) -> dict:
        """
        Returns the core policies in place in the Organization.
        Possible kwargs:
            limit : A positive integer, providing a hint as to the maximum number of resources to return in one page of results.
            property : Filter responses based on a property and optional existence or relational values.
            Only the ‘name’ property is supported for core resources.
            For custom resources, additional supported property values include 'status’, 'created’, 'createdClient’, 'createdUser’, 'updated’, 'updatedClient’, and 'updatedUser’
            orderby : A comma-separated list of properties by which the returned list of resources will be sorted.
            start : Requests items whose ‘orderby’ property value are strictly greater than the supplied ‘start’ value.
            duleLabels : A comma-separated list of DULE labels. Return only those policies whose "deny" expression references any of the labels in this list
            marketingAction : Restrict returned policies to those that reference the given marketing action.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getPoliciesCore")
        path = "/policies/core"
        params = {**kwargs}
        res = self.connector.getData(
            self.endpoint + path, params=params, headers=self.header
        )
        return res

    def getPoliciesCoreId(self, policy_id: str = None):
        """
        Return a specific core policy by its id.
        Arguments:
            policy_id : REQUIRED : policy_id to retrieve.
        """
        if policy_id is None:
            raise Exception("Expected a policy id")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getPoliciesCoreId")
        path = f"/policies/core/{policy_id}"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        return res

    def getPoliciesCustoms(self, **kwargs):
        """
        Returns the custom policies in place in the Organization.
        Possible kwargs:
            limit : A positive integer, providing a hint as to the maximum number of resources to return in one page of results.
            property : Filter responses based on a property and optional existence or relational values.
            Only the ‘name’ property is supported for core resources.
            For custom resources, additional supported property values include 'status’, 'created’, 'createdClient’, 'createdUser’, 'updated’, 'updatedClient’, and 'updatedUser’
            orderby : A comma-separated list of properties by which the returned list of resources will be sorted.
            start : Requests items whose ‘orderby’ property value are strictly greater than the supplied ‘start’ value.
            duleLabels : A comma-separated list of DULE labels. Return only those policies whose "deny" expression references any of the labels in this list
            marketingAction : Restrict returned policies to those that reference the given marketing action.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getPoliciesCustoms")
        path = "/policies/custom"
        params = {**kwargs}
        res = self.connector.getData(
            self.endpoint + path, params=params, headers=self.header
        )
        return res

    def getPoliciesCustom(self, policy_id: str = None):
        """
        Return a specific custom policy by its id.
        Arguments:
            policy_id: REQUIRED: policy_id to retrieve.
        """
        if policy_id is None:
            raise Exception("Expected a policy id")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getPoliciesCustom")
        path = f"/policies/custom/{policy_id}"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        return res

    def createPolicy(self, policy: typing.Union(dict, typing.IO) = None):
        """
        Create a custom policy.
        Arguments:
            policy : REQUIRED : A dictionary contaning the policy you would like to implement.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting createPolicy")
        path = "/policies/custom"
        res = self.connector.postData(self.endpoint + path, data=policy)
        return res

    def getCoreLabels(self, prop: str = None, limit: int = 100) -> list:
        """
        Retrieve a list of core labels.
        Arguments:
            prop : OPTIONAL : Filters responses based on whether a specific property exists, or whose value passes a conditional expression
                Example: prop="name==C1".
                Only the “name” property is supported for core resources.
            limit : OPTIONAL : number of results to be returned. Default 100
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getCoreLabels")
        params = {"limit": limit}
        if prop is not None:
            params["property"] = prop
        path = "/labels/core"
        res = self.connector.getData(
            self.endpoint + path, params=params, headers=self.header
        )
        data = res["children"]
        nextPage = res["_links"]["page"].get("href", "")
        while nextPage != "":
            params["start"] = nextPage
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=self.header
            )
            data += res["children"]
            nextPage = res["_links"]["page"].get("href", "")
        return data

    def getCoreLabel(self, labelName: str = None) -> dict:
        """
        Returns a specific Label by its name.
        Argument:
            labelName : REQUIRED : The name of the core label.
        """
        if labelName is None:
            raise ValueError("Require a label name")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getCoreLabel")
        path = f"/labels/core/{labelName}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def getCustomLabels(self, prop: str = None, limit: int = 100) -> list:
        """
        Retrieve a list of custom labels.
        Arguments:
            prop : OPTIONAL : Filters responses based on whether a specific property exists, or whose value passes a conditional expression
                Example: prop="name==C1".
                Property values include "status", "created", "createdClient", "createdUser", "updated", "updatedClient", and "updatedUser".
            limit : OPTIONAL : number of results to be returned. Default 100
        """
        params = {"limit": limit}
        if prop is not None:
            params["property"] = prop
        if self.loggingEnabled:
            self.logger.debug(f"Starting getCustomLabels")
        path = "/labels/custom"
        res = self.connector.getData(
            self.endpoint + path, params=params, headers=self.header
        )
        data = res["children"]
        nextPage = res["_links"]["page"].get("href", "")
        while nextPage != "":
            params["start"] = nextPage
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=self.header
            )
            data += res["children"]
            nextPage = res["_links"]["page"].get("href", "")
        return data

    def getCustomLabel(self, labelName: str = None) -> dict:
        """
        Returns a specific Label by its name.
        Argument:
            labelName : REQUIRED : The name of the custom label.
        """
        if labelName is None:
            raise ValueError("Require a label name")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getCustomLabel")
        path = f"/labels/custom/{labelName}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def updateCustomLabel(self, labelName: str = None, data: dict = None) -> dict:
        """
        Update a specific Label by its name. (PUT method)
        Argument:
            labelName : REQUIRED : The name of the custom label.
            data : REQUIRED : Data to replace the old definition
                Example:
                {
                    "name": "L2",
                    "category": "Custom",
                    "friendlyName": "Purchase History Data",
                    "description": "Data containing information on past transactions"
                }
        """
        if labelName is None:
            raise ValueError("Require a label name")
        if data is None or type(data) != dict:
            raise ValueError("Require a dictionary data to be passed")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateCustomLabel")
        path = f"/labels/custom/{labelName}"
        res = self.connector.putData(self.endpoint + path, data=data)
        return res

    def getMarketingActionsCore(
        self, prop: str = None, limit: int = 10, **kwargs
    ) -> list:
        """
        Retrieve a list of core marketing actions.
        Arguments:
            prop : OPTIONAL : Filters responses based on whether a specific property exists, or whose value passes a conditional expression (e.g. "prop=name==C1").
            Only the “name” property is supported for core resources.
            limit : OPTIONAL : number of results to be returned.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getMarketingActionsCore")
        path = "/marketingActions/core"
        params = {"limit": limit}
        if prop is not None:
            params["property"] = prop
        res = self.connector.getData(self.endpoint + path, params=params)
        data = res["children"]
        nextPage = res["_links"]["page"].get("href", "")
        while nextPage != "":
            params["start"] = nextPage
            res = self.connector.getData(self.endpoint + path, params=params)
            data += res["children"]
            nextPage = res["_links"]["page"].get("href", "")
        return data



import aepp
from aepp import connector
import logging


class AccessControl:
    """
    Access Control API endpoints.
    Complete documentation is available:
    https://www.adobe.io/apis/experienceplatform/home/api-reference.html#!acpdr/swagger-specs/access-control.yaml
    """

    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(
        self,
        config: dict = aepp.config.config_object,
        header=aepp.config.header,
        loggingObject: dict = None,
        **kwargs,
    ) -> None:
        """
        Instantiate the access controle API wrapper.
        Arguments:
            config : OPTIONAL : config object in the config module. (DO NOT MODIFY)
            header : OPTIONAL : header object  in the config module. (DO NOT MODIFY)
            loggingObject : OPTIONAL : logging object to log messages.
        kwargs :
            header options
        """
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(
            config_object=config,
            header=header,
            loggingEnabled=self.loggingEnabled,
            logger=self.logger,
        )
        self.sandbox = self.connector.config["sandbox"]
        self.header = self.connector.header
        self.header.update(**kwargs)
        self.endpoint = (
            aepp.config.endpoints["global"] + aepp.config.endpoints["access"]
        )

    def getReferences(self) -> dict:
        """
        List all available permission names and resource types.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getReferences")
        path = "/acl/reference"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        return res

    def postEffectivePolicies(self, listElements: list = None):
        """
        List all effective policies for a user on given resources within a sandbox.
        Arguments:
            listElements : REQUIRED : List of resource urls. Example url : /resource-types/{resourceName} or /permissions/{highLevelPermissionName}
        """
        if type(listElements) != list:
            raise TypeError("listElements should be a list of elements")
        if self.loggingEnabled:
            self.logger.debug(f"Starting postEffectivePolicies")
        path = "/acl/effective-policies"
        res = self.connector.postData(
            self.endpoint + path, data=listElements, headers=self.header
        )
        return res



import aepp
from aepp import connector
import logging


class Sandboxes:
    """
    A collection of methods to realize actions on the sandboxes.
    It comes from the sandbox API:
    https://www.adobe.io/apis/experienceplatform/home/api-reference.html#!acpdr/swagger-specs/sandbox-api.yaml
    """

    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(
        self,
        config: dict = aepp.config.config_object,
        header=aepp.config.header,
        loggingObject: dict = None,
        **kwargs,
    ):
        """
        Instantiate the sandbox class.
        Arguments:
            config : OPTIONAL : config object in the config module. (DO NOT MODIFY)
            header : OPTIONAL : header object  in the config module. (DO NOT MODIFY)
            loggingObject : OPTIONAL : logging object to log messages.
        Additional kwargs will update the header.
        """
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(
            config_object=config,
            header=header,
            loggingEnabled=self.loggingEnabled,
            loggingObject=self.logger,
        )
        self.header = self.connector.header
        self.header.update(**kwargs)
        self.endpoint = (
            aepp.config.endpoints["global"] + aepp.config.endpoints["sandboxes"]
        )

    def getSandboxes(self) -> list:
        """
        Return the list of all the sandboxes
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getSandboxes")
        path = self.endpoint + "/sandboxes"
        res = self.connector.getData(path)
        data = res["sandboxes"]
        return data

    def getSandboxTypes(self) -> list:
        """
        Return the list of all the sandboxes types.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getSandboxTyoes")
        path = self.endpoint + "/sandboxTypes"
        res = self.connector.getData(path)
        data = res["sandboxTypes"]
        return data

    def createSandbox(
        self, name: str = None, title: str = None, type_sandbox: str = "development"
    ) -> dict:
        """
        Create a new sandbox in your AEP instance.
        Arguments:
            name : REQUIRED : name of your sandbox
            title : REQUIRED : display name of your sandbox
            type_sandbox : OPTIONAL : type of your sandbox. default : development.
        """
        if name is None or title is None:
            raise Exception("name and title cannot be empty")
        if self.loggingEnabled:
            self.logger.debug(f"Starting createSandbox")
        path = self.endpoint + "/sandboxes"
        data = {"name": name, "title": title, "type": type_sandbox}
        res = self.connector.postData(path, data=data)
        return res

    def getSandbox(self, name: str) -> dict:
        """
        retrieve a Sandbox information by name
        Argument:
            name : REQUIRED : name of Sandbox
        """
        if name is None:
            raise Exception("Expected a name as parameter")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getSandbox")
        path = self.endpoint + f"/sandboxes/{name}"
        res = self.connector.getData(path)
        return res

    def deleteSandbox(self, name: str) -> dict:
        """
        Delete a sandbox by its name.
        Arguments:
            name : REQUIRED : sandbox to be deleted.
        """
        if name is None:
            raise Exception("Expected a name as parameter")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteSandbox")
        path = self.endpoint + f"/sandboxes/{name}"
        res = self.connector.deleteData(path)
        return res

    def resetSandbox(self, name: str) -> dict:
        """
        Reset a sandbox by its name. Sandbox will be empty.
        Arguments:
            name : REQUIRED : sandbox name to be deleted.
        """
        if name is None:
            raise Exception("Expected a sandbox name as parameter")
        if self.loggingEnabled:
            self.logger.debug(f"Starting resetSandbox")
        path = self.endpoint + f"/sandboxes/{name}"
        res = self.connector.putData(path)
        return res

    def updateSandbox(self, name: str, action: dict = None) -> dict:
        """
        Update the Sandbox with the action provided.
        Arguments:
            name : REQUIRED : sandbox name to be updated.
            action : REQUIRED : dictionary defining the action to realize on that sandbox.
        """
        if name is None:
            raise Exception("Expected a sandbox name as parameter")
        if action is None or type(action) != dict:
            raise Exception("Expected a dictionary to pass the action")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateSandboxes")
        path = self.endpoint + f"/sandboxes/{name}"
        res = self.connector.patchData(path, data=action)
        return res



import aepp
from aepp import connector
from copy import deepcopy
import time
import logging


class FlowService:
    """
    The Flow Service manage the ingestion part of the data in AEP.
    For more information, relate to the API Documentation, you can directly refer to the official documentation:
        https://www.adobe.io/apis/experienceplatform/home/api-reference.html#!acpdr/swagger-specs/flow-service.yaml
        https://experienceleague.adobe.com/docs/experience-platform/sources/home.html
        https://experienceleague.adobe.com/docs/experience-platform/destinations/home.html
    """

    PATCH_REFERENCE = [
        {
            "op": "Add",
            "path": "/auth/params",
            "value": {
                "description": "A new description to provide further context on a specified connection or flow."
            },
        }
    ]

    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(
        self,
        config: dict = aepp.config.config_object,
        header=aepp.config.header,
        loggingObject: dict = None,
        **kwargs,
    ):
        """
        initialize the Flow Service instance.
        Arguments:
            config : OPTIONAL : config object in the config module.
            header : OPTIONAL : header object  in the config module.
        """
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(
            config_object=config,
            header=header,
            loggingEnabled=self.loggingEnabled,
            logger=self.logger,
        )
        self.header = self.connector.header
        self.header.update(**kwargs)
        self.sandbox = self.connector.config["sandbox"]
        self.endpoint = aepp.config.endpoints["global"] + aepp.config.endpoints["flow"]

    def getResource(
        self,
        endpoint: str = None,
        params: dict = None,
        format: str = "json",
        save: bool = False,
        **kwargs,
    ) -> dict:
        """
        Template for requesting data with a GET method.
        Arguments:
            endpoint : REQUIRED : The URL to GET
            params: OPTIONAL : dictionary of the params to fetch
            format : OPTIONAL : Type of response returned. Possible values:
                json : default
                txt : text file
                raw : a response object from the requests module
        """
        if endpoint is None:
            raise ValueError("Require an endpoint")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getResource")
        res = self.connector.getData(endpoint, params=params, format=format)
        if save:
            if format == "json":
                aepp.saveFile(
                    module="catalog",
                    file=res,
                    filename=f"resource_{int(time.time())}",
                    type_file="json",
                    encoding=kwargs.get("encoding", "utf-8"),
                )
            elif format == "txt":
                aepp.saveFile(
                    module="catalog",
                    file=res,
                    filename=f"resource_{int(time.time())}",
                    type_file="txt",
                    encoding=kwargs.get("encoding", "utf-8"),
                )
            else:
                print(
                    "element is an object. Output is unclear. No save made.\nPlease save this element manually"
                )
        return res

    def getConnections(
        self, limit: int = 20, n_results: int = 100, count: bool = False, **kwargs
    ) -> list:
        """
        Returns the list of connections available.
        Arguments:
            limit : OPTIONAL : number of result returned per request (default 20)
            n_results : OPTIONAL : number of total result returned (default 100, set to "inf" for retrieving everything)
            count : OPTIONAL : if set to True, just returns the number of connections
        kwargs will be added as query parameters
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getConnections")
        params = {"limit": limit}
        if count:
            params["count"] = count
        for kwarg in kwargs:
            params[kwarg] = kwargs[kwarg]
        path = "/connections"
        res = self.connector.getData(self.endpoint + path, params=params)
        try:
            data = res["items"]
            continuationToken = res.get("_links", {}).get("next", {}).get("href", "")
            while continuationToken != "" and len(data) < float(n_results):
                res = self.connector.getData(
                    self.endpoint + continuationToken, params=params
                )
                data += res["items"]
                continuationToken = (
                    res.get("_links", {}).get("next", {}).get("href", "")
                )
            return data
        except:
            return res

    def createConnection(
        self,
        data: dict = None,
        name: str = None,
        auth: dict = None,
        connectionSpec: dict = None,
        **kwargs,
    ) -> dict:
        """
        Create a connection based on either the data being passed or the information passed.
        Arguments:
            data : REQUIRED : dictionary containing the different elements required for the creation of the connection.

            In case you didn't pass a data parameter, you can pass different information.
            name : REQUIRED : name of the connection.
            auth : REQUIRED : dictionary that contains "specName" and "params"
                specName : string that names of the the type of authentication to be used with the base connection.
                params : dict that contains credentials and values necessary to authenticate and create a connection.
            connectionSpec : REQUIRED : dictionary containing the "id" and "verison" key.
                id : The specific connection specification ID associated with source
                version : Specifies the version of the connection specification ID. Omitting this value will default to the most recent version
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting createConnection")
        path = "/connections"
        if data is not None:
            if (
                "name" not in data.keys()
                or "auth" not in data.keys()
                or "connectionSpec" not in data.keys()
            ):
                raise Exception(
                    "Require some keys to be present : name, auth, connectionSpec"
                )
            obj = data
            res = self.connector.postData(self.endpoint + path, data=obj)
            return res
        elif data is None:
            if "specName" not in auth.keys() or "params" not in auth.keys():
                raise Exception(
                    "Require some keys to be present in auth dict : specName, params"
                )
            if "id" not in connectionSpec.keys():
                raise Exception(
                    "Require some keys to be present in connectionSpec dict : id"
                )
            if name is None:
                raise Exception("Require a name to be present")
            obj = {"name": name, "auth": auth, "connectionSpec": connectionSpec}
            res = self.connector.postData(self.endpoint + path, data=obj)
            return res

    def createStreamingConnection(
        self,
        name: str = None,
        sourceId: str = None,
        dataType: str = "xdm",
        paramName: str = None,
        description: str = "",
        **kwargs,
    ) -> dict:
        """
        Create a Streaming connection
        Arguments:
            name : REQUIRED : Name of the Connection.
            sourceId : REQUIRED : The ID of the streaming connection you want to create.
            dataType : REQUIRED : The type of data to ingest (default xdm)
            paramName : REQUIRED : The name of the streaming connection you want to create.
            description : OPTIONAL : if you want to add a description
        kwargs possibility:
            specName : if you want to modify the specification Name.(Default : "Streaming Connection")
        """
        if name is None:
            raise ValueError("Require a name for the connection")
        if sourceId is None:
            raise Exception("Require an ID for the connection")
        if dataType is None:
            raise Exception("Require a dataType specified")
        if paramName is None:
            raise ValueError("Require a name for the Streaming Connection")
        if self.loggingEnabled:
            self.logger.debug(f"Starting createStreamingConnection")
        obj = {
            "name": name,
            "providerId": "521eee4d-8cbe-4906-bb48-fb6bd4450033",
            "description": description,
            "connectionSpec": {
                "id": "bc7b00d6-623a-4dfc-9fdb-f1240aeadaeb",
                "version": "1.0",
            },
            "auth": {
                "specName": kwargs.get("specName", "Streaming Connection"),
                "params": {
                    "sourceId": sourceId,
                    "dataType": dataType,
                    "name": paramName,
                },
            },
        }
        res = self.createConnection(data=obj)
        return res

    def getConnection(self, connectionId: str = None) -> dict:
        """
        Returns a specific connection object.
        Argument:
            connectionId : REQUIRED : The ID of the connection you wish to retrieve.
        """
        if connectionId is None:
            raise Exception("Require a connectionId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getConnection")
        path = f"/connections/{connectionId}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def connectionTest(self, connectionId: str = None) -> dict:
        """
        Test a specific connection ID.
        Argument:
            connectionId : REQUIRED : The ID of the connection you wish to test.
        """
        if connectionId is None:
            raise Exception("Require a connectionId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting connectionTest")
        path: str = f"/connections/{connectionId}/test"
        res: dict = self.connector.getData(self.endpoint + path)
        return res

    def deleteConnection(self, connectionId: str = None) -> dict:
        """
        Delete a specific connection ID.
        Argument:
            connectionId : REQUIRED : The ID of the connection you wish to delete.
        """
        if connectionId is None:
            raise Exception("Require a connectionId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteConnection")
        path: str = f"/connections/{connectionId}"
        res: dict = self.connector.deleteData(self.endpoint + path)
        return res

    def getConnectionSpecs(self) -> list:
        """
        Returns the list of connectionSpecs in that instance.
        If that doesn't work, return the response.
        """
        path: str = "/connectionSpecs"
        if self.loggingEnabled:
            self.logger.debug(f"Starting getConnectionSpecs")
        res: dict = self.connector.getData(self.endpoint + path)
        try:
            data: list = res["items"]
            return data
        except:
            return res

    def getConnectionSpec(self, specId: str = None) -> dict:
        """
        Returns the detail for a specific connection.
        Arguments:
            specId : REQUIRED : The specification ID of a connection
        """
        if specId is None:
            raise Exception("Require a specId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getConnectionSpec")
        path: str = f"/connectionSpecs/{specId}"
        res: dict = self.connector.getData(self.endpoint + path)
        return res

    def getFlows(
        self,
        limit: int = 10,
        n_results: int = 100,
        prop: str = None,
        filterMappingSetIds: list = None,
        filterSourceIds: list = None,
        filterTargetIds: list = None,
        **kwargs,
    ) -> list:
        """
        Returns the flows set between Source and Target connection.
        Arguments:
            limit : OPTIONAL : number of results returned
            n_results : OPTIONAL : total number of results returned (default 100, set to "inf" for retrieving everything)
            prop : OPTIONAL : comma separated list of top-level object properties to be returned in the response.
                Used to cut down the amount of data returned in the response body.
                For example, prop=id==3416976c-a9ca-4bba-901a-1f08f66978ff,6a8d82bc-1caf-45d1-908d-cadabc9d63a6,3c9b37f8-13a6-43d8-bad3-b863b941fedd.
            filterMappingSetId : OPTIONAL : returns only the flow that possess the mappingSetId passed in a list.
            filterSourceIds : OPTIONAL : returns only the flow that possess the sourceConnectionIds passed in a list.
            filterTargetIds : OPTIONAL : returns only the flow that possess the targetConnectionIds passed in a list.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getFlows")
        params: dict = {"limit": limit, "count": kwargs.get("count", False)}
        if property is not None:
            params["property"] = prop
        if kwargs.get("continuationToken", False) != False:
            params["continuationToken"] = kwargs.get("continuationToken")
        path: str = "/flows"
        res: dict = self.connector.getData(self.endpoint + path, params=params)
        token: str = res.get("_links", {}).get("next", {}).get("href", "")
        items = res["items"]
        while token != "" and len(items) < float(n_results):
            continuationToken = token.split("=")[1]
            params["continuationToken"] = continuationToken
            res = self.connector.getData(self.endpoint + path, params=params)
            token = res["_links"].get("next", {}).get("href", "")
            items += res["items"]
        if filterMappingSetIds is not None:
            filteredItems = []
            for mappingsetId in filterMappingSetIds:
                for item in items:
                    if "transformations" in item.keys():
                        for element in item["transformations"]:
                            if element["params"].get("mappingId", "") == mappingsetId:
                                filteredItems.append(item)
            items = filteredItems
        if filterSourceIds is not None:
            filteredItems = []
            for sourceId in filterSourceIds:
                for item in items:
                    if sourceId in item["sourceConnectionIds"]:
                        filteredItems.append(item)
            items = filteredItems
        if filterTargetIds is not None:
            filteredItems = []
            for targetId in filterTargetIds:
                for item in items:
                    if targetId in item["targetConnectionIds"]:
                        filteredItems.append(item)
            items = filteredItems
        return items

    def getFlow(self, flowId: str = None) -> dict:
        """
        Returns the details of a specific flow.
        Arguments:
            flowId : REQUIRED : the flow ID to be returned
        """
        if flowId is None:
            raise Exception("Require a flowId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getFlow")
        path: str = f"/flows/{flowId}"
        res: dict = self.connector.getData(self.endpoint + path)
        return res

    def deleteFlow(self, flowId: str = None) -> dict:
        """
        Delete a specific flow by its ID.
        Arguments:
            flowId : REQUIRED : the flow ID to be returned
        """
        if flowId is None:
            raise Exception("Require a flowId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteFlow")
        path: str = f"/flows/{flowId}"
        res: dict = self.connector.deleteData(self.endpoint + path)
        return res

    def createFlow(self, obj: dict = None) -> dict:
        """
        Create a flow with the API.
        Arguments:
            obj : REQUIRED : body to create the flow service.
                Details can be seen at https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Flows/postFlow
                requires following keys : name, flowSpec, sourceConnectionIds, targetConnectionIds, transformations, scheduleParams.
        """
        if obj is None:
            raise Exception("Require a dictionary to create the flow")
        if "name" not in obj.keys():
            raise KeyError("missing 'name' parameter in the dictionary")
        if "flowSpec" not in obj.keys():
            raise KeyError("missing 'flowSpec' parameter in the dictionary")
        if "sourceConnectionIds" not in obj.keys():
            raise KeyError("missing 'sourceConnectionIds' parameter in the dictionary")
        if "targetConnectionIds" not in obj.keys():
            raise KeyError("missing 'targetConnectionIds' parameter in the dictionary")
        if self.loggingEnabled:
            self.logger.debug(f"Starting createFlow")
        path: str = "/flows"
        res: dict = self.connector.postData(self.endpoint + path, data=obj)
        return res

    def updateFlow(
        self, flowId: str = None, etag: str = None, updateObj: list = None
    ) -> dict:
        """
        update the flow based on the operation provided.
        Arguments:
            flowId : REQUIRED : the ID of the flow to Patch.
            etag : REQUIRED : ETAG value for patching the Flow.
            updateObj : REQUIRED : List of operation to realize on the flow.

            Follow the following structure:
            [
                {
                    "op": "Add",
                    "path": "/auth/params",
                    "value": {
                    "description": "A new description to provide further context on a specified connection or flow."
                    }
                }
            ]
        """
        if flowId is None:
            raise Exception("Require a flow ID to be present")
        if etag is None:
            raise Exception("Require etag to be present")
        if updateObj is None:
            raise Exception("Require a list with data to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateFlow")
        privateHeader = deepcopy(self.header)
        privateHeader["if-match"] = etag
        path: str = f"/flows/{flowId}"
        res: dict = self.connector.patchData(
            self.endpoint + path, headers=privateHeader, data=updateObj
        )
        return res

    def getFlowSpecs(self, prop: str = None) -> list:
        """
        Returns the flow specifications.
        Arguments:
            prop : OPTIONAL : A comma separated list of top-level object properties to be returned in the response.
                Used to cut down the amount of data returned in the response body.
                For example, prop=id==3416976c-a9ca-4bba-901a-1f08f66978ff,6a8d82bc-1caf-45d1-908d-cadabc9d63a6,3c9b37f8-13a6-43d8-bad3-b863b941fedd.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getFlowSpecs")
        path: str = "/flowSpecs"
        params = {}
        if prop is not None:
            params["property"] = prop
        res: dict = self.connector.getData(self.endpoint + path, params=params)
        items: list = res["items"]
        return items

    def getFlowSpec(self, flowSpecId) -> dict:
        """
        Return the detail of a specific flow ID Spec
        Arguments:
            flowSpecId : REQUIRED : The flow ID spec to be checked
        """
        if flowSpecId is None:
            raise Exception("Require a flowSpecId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getFlowSpec")
        path: str = f"/flowSpecs/{flowSpecId}"
        res: dict = self.connector.getData(self.endpoint + path)
        return res

    def getRuns(
        self, limit: int = 10, n_results: int = 100, prop: str = None, **kwargs
    ) -> list:
        """
        Returns the list of runs. Runs are instances of a flow execution.
        Arguments:
            limit : OPTIONAL : number of results returned per request
            n_results : OPTIONAL : total number of results returned (default 100, set to "inf" for retrieving everything)
            prop : OPTIONAL : comma separated list of top-level object properties to be returned in the response.
                Used to cut down the amount of data returned in the response body.
                For example, prop=id==3416976c-a9ca-4bba-901a-1f08f66978ff,6a8d82bc-1caf-45d1-908d-cadabc9d63a6,3c9b37f8-13a6-43d8-bad3-b863b941fedd.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getRuns")
        path = "/runs"
        params = {"limit": limit, "count": kwargs.get("count", False)}
        if prop is not None:
            params["property"] = prop
        if kwargs.get("continuationToken", False):
            params["continuationToken"] = kwargs.get("continuationToken")
        res: dict = self.connector.getData(self.endpoint + path, params=params)
        items: list = res["items"]
        if res["_links"].get("next", {}).get("href", "") != "" and len(items) < float(
            n_results
        ):
            token: str = res["_links"]["next"].get("href", "")
            continuationToken: str = token.split("=")[1]
            params["continuationToken"] = continuationToken
            items += self.connector.getData(self.endpoint + path, params=params)
        return items

    def createRun(self, flowId: str = None, status: str = "active") -> dict:
        """
        Generate a run based on the flowId.
        Arguments:
            flowId : REQUIRED : the flow ID to run
            stats : OPTIONAL : Status of the flow
        """
        path = "/runs"
        if flowId is None:
            raise Exception("Require a flowId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting createRun")
        obj = {"flowId": flowId, "status": status}
        res: dict = self.connector.postData(self.endpoint + path, data=obj)
        return res

    def getRun(self, runId: str = None) -> dict:
        """
        Return a specific runId.
        Arguments:
            runId : REQUIRED : the run ID to return
        """
        if runId is None:
            raise Exception("Require a runId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getRun")
        path: str = f"/runs/{runId}"
        res: dict = self.connector.getData(self.endpoint + path)
        return res

    def getSourceConnections(self, n_results: int = 100, **kwargs) -> list:
        """
        Return the list of source connections
        Arguments:
            n_results : OPTIONAL : total number of results returned (default 100, set to "inf" for retrieving everything)
        kwargs will be added as query parameterss
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getSourceConnections")
        params = {**kwargs}
        path: str = f"/sourceConnections"
        res: dict = self.connector.getData(self.endpoint + path, params=params)
        data: list = res["items"]
        nextPage = res["_links"].get("next", {}).get("href", "")
        while nextPage != "" and len(data) < float(n_results):
            continuationToken = nextPage.split("=")[1]
            params["continuationToken"] = continuationToken
            res: dict = self.connector.getData(self.endpoint + path, params=params)
            data += res["items"]
            nextPage = res["_links"].get("next", {}).get("href", "")
        return data

    def getSourceConnection(self, sourceConnectionId: str = None) -> dict:
        """
        Return detail of the sourceConnection ID
        Arguments:
            sourceConnectionId : REQUIRED : The source connection ID to be retrieved
        """
        if sourceConnectionId is None:
            raise Exception("Require a sourceConnectionId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getSourceConnection")
        path: str = f"/sourceConnections/{sourceConnectionId}"
        res: dict = self.connector.getData(self.endpoint + path)
        return res

    def deleteSourceConnection(self, sourceConnectionId: str = None) -> dict:
        """
        Delete a sourceConnection ID
        Arguments:
            sourceConnectionId : REQUIRED : The source connection ID to be deleted
        """
        if sourceConnectionId is None:
            raise Exception("Require a sourceConnectionId to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteSourceConnection")
        path: str = f"/sourceConnections/{sourceConnectionId}"
        res: dict = self.connector.deleteData(self.endpoint + path)
        return res

    def createSourceConnection(self, data: dict = None) -> dict:
        """
        Create a sourceConnection based on the dictionary passed.
        Arguments:
            obj : REQUIRED : the data to be passed for creation of the Source Connection.
                Details can be seen at https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Source_connections/postSourceConnection
                requires following keys : name, baseConnectionId, data, params, connectionSpec.
        """
        if data is None:
            raise Exception("Require a dictionary with data to be present")
        if "name" not in data.keys():
            raise KeyError("Require a 'name' key in the dictionary passed")
        if "connectionSpec" not in data.keys():
            raise KeyError("Require a 'connectionSpec' key in the dictionary passed")
        if self.loggingEnabled:
            self.logger.debug(f"Starting createSourceConnection")
        path: str = f"/sourceConnections"
        res: dict = self.connector.postData(self.endpoint + path, data=data)
        return res

    def createSourceConnectionStreaming(
        self,
        connectionId: str = None,
        name: str = None,
        format: str = "delimited",
        description: str = "",
    ) -> dict:
        """
        Create a source connection based on streaming connection created.
        Arguments:
            connectionId : REQUIRED : The Streaming connection ID.
            name : REQUIRED : Name of the Connection.
            format : REQUIRED : format of the data sent (default : delimited)
            description : OPTIONAL : Description of of the Connection Source.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting createSourceConnectionStreaming")
        obj = {
            "name": name,
            "providerId": "521eee4d-8cbe-4906-bb48-fb6bd4450033",
            "description": description,
            "baseConnectionId": connectionId,
            "connectionSpec": {
                "id": "bc7b00d6-623a-4dfc-9fdb-f1240aeadaeb",
                "version": "1.0",
            },
            "data": {"format": format},
        }
        res = self.createSourceConnection(data=obj)
        return res

    def updateSourceConnection(
        self, sourceConnectionId: str = None, etag: str = None, updateObj: list = None
    ) -> dict:
        """
        Update a source connection based on the ID provided with the object provided.
        Arguments:
            sourceConnectionId : REQUIRED : The source connection ID to be updated
            etag: REQUIRED : A header containing the etag value of the connection or flow to be updated.
            updateObj : REQUIRED : The operation call used to define the action needed to update the connection. Operations include add, replace, and remove.
        """
        if sourceConnectionId is None:
            raise Exception("Require a sourceConnection to be present")
        if etag is None:
            raise Exception("Require etag to be present")
        if updateObj is None:
            raise Exception("Require a list with data to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateSourceConnection")
        privateHeader = deepcopy(self.header)
        privateHeader["if-match"] = etag
        path: str = f"/sourceConnections/{sourceConnectionId}"
        res: dict = self.connector.patchData(
            self.endpoint + path, headers=privateHeader, data=updateObj
        )
        return res

    def getTargetConnections(self, n_results: int = 100, **kwargs) -> dict:
        """
        Return the target connections
        Arguments:
            n_results : OPTIONAL : total number of results returned (default 100, set to "inf" for retrieving everything)
        kwargs will be added as query parameterss
        """
        params = {**kwargs}
        path: str = f"/targetConnections"
        res: dict = self.connector.getData(self.endpoint + path, params=params)
        data: list = res["items"]
        nextPage = res["_links"].get("next", {}).get("href", "")
        while nextPage != "" and len(data) < float(n_results):
            continuationToken = nextPage.split("=")[1]
            params["continuationToken"] = continuationToken
            res: dict = self.connector.getData(self.endpoint + path, params=params)
            data += res["items"]
            nextPage = res["_links"].get("next", {}).get("href", "")
        return data

    def getTargetConnection(self, targetConnectionId: str = None) -> dict:
        """
        Retrieve a specific Target connection detail.
        Arguments:
            targetConnectionId : REQUIRED : The target connection ID is a unique identifier used to create a flow.
        """
        if targetConnectionId is None:
            raise Exception("Require a target connection ID to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getTargetConnection")
        path: str = f"/targetConnections/{targetConnectionId}"
        res: dict = self.connector.getData(self.endpoint + path)
        return res

    def deleteTargetConnection(self, targetConnectionId: str = None) -> dict:
        """
        Delete a specific Target connection detail
        Arguments:
             targetConnectionId : REQUIRED : The target connection ID to be deleted
        """
        if targetConnectionId is None:
            raise Exception("Require a target connection ID to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteTargetConnection")
        path: str = f"/targetConnections/{targetConnectionId}"
        res: dict = self.connector.deleteData(self.endpoint + path)
        return res

    def createTargetConnection(
        self,
        name: str = None,
        connectionSpecId: str = None,
        datasetId: str = None,
        format: str = "parquet_xdm",
        version: str = "1.0",
        description: str = "",
        data: dict = None,
    ) -> dict:
        """
        Create a new target connection
        Arguments:
                name : REQUIRED : The name of the target connection
                connectionSpecId : REQUIRED : The connectionSpecId to use.
                datasetId : REQUIRED : The dataset ID that is the target
                version : REQUIRED : version to be used (1.0 by default)
                format : REQUIRED : Data format to be used (parquet_xdm by default)
                description : OPTIONAL : description of your target connection
                data : OPTIONAL : If you pass the complete dictionary for creation
        Details can be seen at https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Target_connections/postTargetConnection
        requires following keys : name, data, params, connectionSpec.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting createTargetConnection")
        path: str = f"/targetConnections"
        if data is not None and type(data) == dict:
            obj = data
            res: dict = self.connector.postData(self.endpoint + path, data=obj)
        else:
            if name is None:
                raise ValueError("Require a name to be passed")
            if connectionSpecId is None:
                raise ValueError("Require a connectionSpec Id to be passed")
            if datasetId is None:
                raise ValueError("Require a datasetId to be passed")
            obj = {
                "name": name,
                "description": description,
                "connectionSpec": {"id": connectionSpecId, "version": version},
                "data": {"format": format},
                "params": {"dataSetId": datasetId},
            }
            res: dict = self.connector.postData(self.endpoint + path, data=obj)
        return res

    def createTargetConnectionDataLake(
        self,
        name: str = None,
        datasetId: str = None,
        schemaId: str = None,
        format: str = "delimited",
        version: str = "1.0",
        description: str = "",
    ) -> dict:
        """
        Create a target connection to the AEP Data Lake.
        Arguments:
            name : REQUIRED : The name of your target Destination
            datasetId : REQUIRED : the dataset ID of your target destination.
            schemaId : REQUIRED : The schema ID of your dataSet. (NOT meta:altId)
            format : REQUIRED : format of your data inserted
            version : REQUIRED : version of your target destination
            description : OPTIONAL : description of your target destination.
        """
        targetObj = {
            "name": name,
            "description": description,
            "data": {
                "format": format,
                "schema": {
                    "id": schemaId,
                    "version": "application/vnd.adobe.xed-full+json;version=1.0",
                },
            },
            "params": {"dataSetId": datasetId},
            "connectionSpec": {
                "id": "c604ff05-7f1a-43c0-8e18-33bf874cb11c",
                "version": version,
            },
        }
        if self.loggingEnabled:
            self.logger.debug(f"Starting createTargetConnectionDataLake")
        res = self.createTargetConnection(data=targetObj)
        return res

    def updateTargetConnection(
        self, targetConnectionId: str = None, etag: str = None, updateObj: list = None
    ) -> dict:
        """
        Update a target connection based on the ID provided with the object provided.
        Arguments:
            targetConnectionId : REQUIRED : The target connection ID to be updated
            etag: REQUIRED : A header containing the etag value of the connection or flow to be updated.
            updateObj : REQUIRED : The operation call used to define the action needed to update the connection. Operations include add, replace, and remove.
        """
        if targetConnectionId is None:
            raise Exception("Require a sourceConnection to be present")
        if etag is None:
            raise Exception("Require etag to be present")
        if updateObj is None:
            raise Exception("Require a dictionary with data to be present")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateTargetConnection")
        privateHeader = deepcopy(self.header)
        privateHeader["if-match"] = etag
        path: str = f"/targetConnections/{targetConnectionId}"
        res: dict = self.connector.patchData(
            self.endpoint + path, headers=privateHeader, data=updateObj
        )
        return res



# Internal Library
from aepp import config
from aepp import connector
import pandas as pd
from pg import DB
from typing import Union
import re
import aepp
import time
import logging


class QueryService:
    """
    QueryService class is to be used in order to generate queries, scheduled queries or retrieve past queries.
    """

    QUERYSAMPLE = {
        "dbName": "string",
        "sql": "SELECT $key from $key1 where $key > $key2;",
        "queryParameters": {"key": "value", "key1": "value1", "key2": "value2"},
        "templateId": "123",
        "name": "string",
        "description": "powered by aepp",
        "insertIntoParameters": {"datasetName": "string"},
        "ctasParameters": {
            "datasetName": "myDatasetName",
            "description": "powered by aepp",
            "targetSchemaTitle": "mySchemaName",
        },
    }
    SCHEDULESAMPLE = {
        "query": {
            "dbName": "string",
            "sql": "SELECT $key from $key1 where $key > $key2;",
            "queryParameters": {"key": "value", "key1": "value1", "key2": "value2"},
            "templateId": "123",
            "name": "string",
            "description": "string",
            "insertIntoParameters": {"datasetName": "string"},
            "ctasParameters": {
                "datasetName": "string",
                "description": "string",
                "targetSchemaTitle": "mySchemaName",
            },
        },
        "schedule": {
            "schedule": "string",
            "startDate": "string",
            "endDate": "string",
            "maxActiveRuns": 0,
        },
    }
    TEMPLATESAMPLE = {
        "sql": "SELECT $key from $key1 where $key > $key2;",
        "queryParameters": {"key": "value", "key1": "value1", "key2": "value2"},
        "name": "string",
    }

    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(
        self,
        config_object: dict = config.config_object,
        header=config.header,
        loggingObject: dict = None,
        **kwargs,
    ) -> None:
        """
        Instanciate the class for Query Service call.
        Arguments:
            config_object : OPTIONAL : config object in the config module. (DO NOT MODIFY)
            header : OPTIONAL : header object  in the config module. (DO NOT MODIFY)
            loggingObject : OPTIONAL : If you want to set logging capability for your actions.
        kwargs:
            kwargs will update the header
        """
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(
            config_object=config_object,
            header=header,
            loggingEnabled=self.loggingEnabled,
            logger=self.logger,
        )
        self.header = self.connector.header
        # self.header.update({"Accept": "application/json"})
        self.header.update(**kwargs)
        self.sandbox = self.connector.config["sandbox"]
        self.endpoint = config.endpoints["global"] + config.endpoints["query"]

    def getResource(
        self,
        endpoint: str = None,
        params: dict = None,
        format: str = "json",
        save: bool = False,
        **kwargs,
    ) -> dict:
        """
        Template for requesting data with a GET method.
        Arguments:
            endpoint : REQUIRED : The URL to GET
            params: OPTIONAL : dictionary of the params to fetch
            format : OPTIONAL : Type of response returned. Possible values:
                json : default
                txt : text file
                raw : a response object from the requests module
        """
        if endpoint is None:
            raise ValueError("Require an endpoint")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getResource")
        res = self.connector.getData(endpoint, params=params, format=format)
        if save:
            if format == "json":
                aepp.saveFile(
                    module="catalog",
                    file=res,
                    filename=f"resource_{int(time.time())}",
                    type_file="json",
                    encoding=kwargs.get("encoding", "utf-8"),
                )
            elif format == "txt":
                aepp.saveFile(
                    module="catalog",
                    file=res,
                    filename=f"resource_{int(time.time())}",
                    type_file="txt",
                    encoding=kwargs.get("encoding", "utf-8"),
                )
            else:
                print(
                    "element is an object. Output is unclear. No save made.\nPlease save this element manually"
                )
        return res

    def connection(self) -> dict:
        """
        Create a connection for interactive interface.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Getting a connection")
        path = "/connection_parameters"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        return res

    def getQueries(
        self,
        orderby: str = None,
        limit: int = 1000,
        start: int = None,
        n_results: int = 1000,
        property: str = None,
        **kwargs,
    ) -> list:
        """
        Retrieve the queries from your organization.
        Arguments:
            orderby : OPTIONAL : possibility to order by "created" or "updated".
            prepend with "+" for ASC and "-" for DESC. To be escaped (default: -)
            limit : OPTIONAL : number of of records to fetch per page. (default 1000)
            start : OPTIONAL : when to start (depending on your orderby)
            property : OPTIONAL : Comma-separated filters. List of properties that allow filtering with all operators:
                        "created"
                        "updated"
                        "state"
                        "id"
                    with following operators >, <, >=, <=, ==, !=, ~
                "referenced_datasets", "scheduleId", "templateId", and "userId" can only be used with ==. Multiple datasets ID with "||"
                "SQL" can only be used with ~ (WITHOUT COMMA)
            n_results : OPTIONAL : total number of results returned (default 1000, set to "inf" to retrieve everything)
        possible kwargs:
            excludeSoftDeleted: Whether to include any queries that have been soft deleted. Defaults to true.
            excludeHidden : Whether to include any queries that have been found to be not interesting, as they were not user driven. Examples include CURSOR definitions, FETCH, and Metadata queries. Defaults to true.
            isPrevLink : Field indicating if the URI is a previous link.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getQueries")
        path = "/queries"
        arguments = {"limit": 1000}
        n_results = float(n_results)
        if orderby is not None:
            if orderby == "+":
                orderby = "%2B"
            arguments["orderby"] = orderby
        if start is not None:
            arguments["start"] = start
        if limit is not None:
            arguments["limit"] = limit
        if property is not None:
            arguments["property"] = property
        if len(kwargs.keys()) > 0:
            arguments["excludeSoftDeleted"] = kwargs.get("excludeSoftDeleted", True)
            arguments["excludeHidden"] = kwargs.get("excludeHidden", True)
            arguments["isPrevLink"] = kwargs.get("isPrevLink", "")
        res = self.connector.getData(self.endpoint + path, params=arguments)
        data = res["queries"]
        nextPage = res["_links"].get("next", {}).get("href", "")
        while nextPage != "":
            hrefParams = nextPage.split("?")[1]
            orderBy = re.search("orderby=(.+?)(&|$)", hrefParams)
            start = re.search("start=(.+?)(&|$)", hrefParams)
            arguments["start"] = start.group(1)
            arguments["orderby"] = orderBy.group(1)
            res = self.connector.getData(self.endpoint + path, params=arguments)
            data += res.get("queries", [])
            nextPage = res.get("_links", {}).get("next", {}).get("href", "")
            if n_results < float(
                len(data)
            ):  ## forcing exit when reaching number of results asked
                nextPage = ""
        return data

    def postQueries(
        self,
        data: dict = None,
        name: str = None,
        dbname: str = "prod:all",
        sql: str = None,
        templateId: str = None,
        queryParameters: dict = None,
        insertIntoParameters: dict = None,
        ctasParameters: dict = None,
        description: str = "",
        **kwargs,
    ) -> dict:
        """
        Create a query.
        Arguments:
            data : OPTIONAL : If you want to pass the full query statement.
            name : REQUIRED : Name of the query
            dbname : REQUIRED : the dataset name (default prod:all)
            sql: REQUIRED : the SQL query as a string.
            queryParameters : OPTIONAL : in case you are using template, providing the paramter in a dictionary.
            insertIntoParameters : OPTIONAL : in case you want to insert the result to an existing dataset
                example : {
                    "datasetName": "string"
                }
            ctasParameters: OPTIONAL : in case you want to create a dataset out of that query, dictionary is required with "datasetName" and "description".
                example : {
                    "datasetName": "string",
                    "description": "string",
                    "targetSchemaTitle":"string"
                    }
                    targetSchemaTitle if you want to use a precreated schema.
        """
        path = "/queries"
        if self.loggingEnabled:
            self.logger.debug(f"Starting postQuery")
        if data is None:
            if sql is None or name is None:
                raise AttributeError("You are missing required arguments.")
            if type(ctasParameters) is dict:
                if (
                    "datasetName" not in ctasParameters.keys()
                    or "description" not in ctasParameters.keys()
                ):
                    raise KeyError(
                        'Expecting "datasetName" and "description" as part of the the ctasParameters dictionary.'
                    )
            data = {
                "name": name,
                "description": description,
                "dbName": dbname,
                "sql": sql,
            }
            if templateId is not None:
                data["templateId"] = templateId
            if queryParameters is not None:
                data["queryParameters"] = queryParameters
            if ctasParameters is not None:
                data["ctasParameters"] = ctasParameters
            if insertIntoParameters is not None:
                data["insertIntoParameters"] = insertIntoParameters
        res = self.connector.postData(
            self.endpoint + path, data=data, headers=self.header
        )
        return res

    def getQuery(self, queryId: str = None) -> dict:
        """
        Request the query status by ID.
        Argument :
            queryid : REQUIRED : the query id to check
        """
        if queryId is None:
            raise AttributeError('Expected "queryId" to be filled')
        if self.loggingEnabled:
            self.logger.debug(f"Starting getQuery")
        path = f"/queries/{queryId}"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        return res

    def cancelQuery(self, queryId: str = None) -> dict:
        """
        Cancel a specific query based on its ID.
        Argument:
            queryId : REQUIRED : the query ID to cancel
        """
        if queryId is None:
            raise AttributeError('Expected "queryId" to be filled')
        if self.loggingEnabled:
            self.logger.debug(f"Starting cancelQuery")
        path = f"/queries/{queryId}"
        data = {"op": "cancel"}
        res = self.connector.patchData(self.endpoint + path, data=data)
        return res

    def deleteQuery(self, queryId: str = None) -> dict:
        """
        Delete a specific query based on its ID.
        Argument:
            queryId : REQUIRED : the query ID to delete
        """
        if queryId is None:
            raise AttributeError('Expected "queryId" to be filled')
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteQuery")
        path = f"/queries/{queryId}"
        res = self.connector.deleteData(self.endpoint + path)
        return res

    def getSchedules(self, n_results: int = 1000, **kwargs) -> list:
        """
        Retrieve a list of scheduled queries for the AEP instance.
        Arguments:
            n_results : OPTIONAL : Total number of scheduled queries retrieved. To get them all, use "inf"
        possibile kwargs:
            orderby : + for ASC and - for DESC
            limit : number of record to fetch
            start : specific start (use with orderby)
            property : Comma-separated filters.
                created ; ex : created>2017-04-05T13:30:00Z
                templateId ; ex : templateId==123412354
                userId ; ex : userId==12341235
        """
        n_results = float(n_results)
        if kwargs.get("orderby", None) is not None:
            if kwargs.get("orderby", "-") == "+":
                kwargs.get("orderby", "-") == "%2B"
        if self.loggingEnabled:
            self.logger.debug(f"Starting getSchedules")
        path = "/schedules"
        params = {**kwargs}
        res = self.connector.getData(self.endpoint + path, params=params)
        data = res["schedules"]
        nextPage = res["_links"].get("next", {}).get("href", "")
        while nextPage != "":
            hrefParams = nextPage.split("?")[1]
            orderBy = re.search("orderby=(.+?)(&|$)", hrefParams)
            start = re.search("start=(.+?)(&|$)", hrefParams)
            params["start"] = start.group(1)
            params["orderby"] = orderBy.group(1)
            res = self.connector.getData(self.endpoint + path, params=params)
            data += res.get("schedules", [])
            nextPage = res["_links"].get("next", {}).get("href", "")
            if len(data) > float(n_results):
                nextPage = ""
        return data

    def getSchedule(self, scheduleId: str = None) -> dict:
        """
        Get a details about a schedule query.
        Arguments:
            scheduleId : REQUIRED : the schedule id
        """
        if scheduleId is None:
            raise Exception("scheduleId is required")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getSchedule")
        path = f"/schedules/{scheduleId}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def getScheduleRuns(self, scheduleId: str = None) -> list:
        """
        Get the different jobs ran for this schedule
        Arguments:
            scheduleId : REQUIRED : the schedule id
        """
        if scheduleId is None:
            raise Exception("scheduleId is required")
        path = f"/schedules/{scheduleId}/runs"
        if self.loggingEnabled:
            self.logger.debug(f"Starting getScheduleRuns")
        params = {}
        res = self.connector.getData(self.endpoint + path)
        data = res["runsSchedules"]
        nextPage = res["_links"].get("next", {}).get("href", "")
        while nextPage != "":
            hrefParams = nextPage.split("?")[1]
            orderBy = re.search("orderby=(.+?)(&|$)", hrefParams)
            start = re.search("start=(.+?)(&|$)", hrefParams)
            params["start"] = start.group(1)
            params["orderby"] = orderBy.group(1)
            res = self.connector.getData(self.endpoint + path, params=params)
            data += res.get("runsSchedules", [])
            nextPage = res["_links"].get("next", {}).get("href", "")
        return data

    def getScheduleRun(self, scheduleId: str = None, runId: str = None) -> dict:
        """
        Get the different jobs ran for this schedule
        Arguments:
            scheduleId : REQUIRED : the schedule id
            runId : REQUIRED : the run ID you want to retrieve.
        """
        if scheduleId is None or runId is None:
            raise Exception("scheduleId and jobId are required")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getScheduleRun")
        path = f"/schedules/{scheduleId}/runs/{runId}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def createSchedule(
        self,
        scheduleQuery: dict = None,
        name: str = None,
        dbname: str = "prod:all",
        sql: str = None,
        templateId: str = None,
        queryParameters: dict = None,
        insertIntoParameters: dict = None,
        ctasParameters: dict = None,
        schedule: dict = None,
        description: str = "",
        **kwargs,
    ) -> dict:
        """
        Create a scheduled query.
        Arguments:
            scheduleQuery: REQUIRED : a dictionary containing the query and the schedule.
            name : OPTIONAL : Name of the query
            dbname : OPTIONAL : the dataset name (default prod:all)
            sql: OPTIONAL : the SQL query as a string.
            queryParameters : OPTIONAL : in case you are using template, providing the paramter in a dictionary.
            ctasParameters: OPTIONAL : in case you want to create a dataset out of that query, dictionary is required with "datasetName" and "description".
            schedule : OPTIONAL : Dictionary giving the instruction to schedule the query.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting createSchedule")
        if scheduleQuery is None:
            if name is None or sql is None or schedule is None:
                raise Exception(
                    "Expecting either scheduleQUery dictionary or data in parameters"
                )
            scheduleQuery = {
                "query": {
                    "name": name,
                    "description": description,
                    "dbName": dbname,
                    "sql": sql,
                },
                "schedule": schedule,
            }
            if templateId is not None:
                scheduleQuery["query"]["templateId"] = templateId
            if queryParameters is not None:
                scheduleQuery["query"]["queryParameters"] = queryParameters
            if ctasParameters is not None:
                scheduleQuery["query"]["ctasParameters"] = ctasParameters
            if insertIntoParameters is not None:
                scheduleQuery["query"]["insertIntoParameters"] = insertIntoParameters
        if type(scheduleQuery) != dict:
            raise Exception("scheduleQuery is required and should be dictionary. ")
        if (
            "query" not in scheduleQuery.keys()
            or "schedule" not in scheduleQuery.keys()
        ):
            raise Exception(
                'scheduleQuery dictonary is expected to contain "schedule" or "query" keys.'
            )
        path = "/schedules"
        res = self.connector.postData(self.endpoint + path, data=scheduleQuery)
        return res

    def deleteSchedule(self, scheduleId: str = None) -> Union[str, dict]:
        """
        Delete a scheduled query.
        Arguments:
            scheduleId : REQUIRED : id of the schedule.
        """
        if scheduleId is None:
            raise Exception("Missing scheduleId")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteSchedules")
        path = f"/schedules/{scheduleId}"
        res = self.connector.deleteData(self.endpoint + path)
        return res

    def disableSchedule(self, scheduleId: str = None) -> dict:
        """
        Disable a scheduled query.
        Arguments:
            scheduleId : REQUIRED : id of the schedule.
        """
        if scheduleId is None:
            raise Exception("Missing scheduleId")
        if self.loggingEnabled:
            self.logger.debug(f"Starting disableSchedules")
        obj = {"body": [{"op": "replace", "path": "/state", "value": "disable"}]}
        path = f"/schedules/{scheduleId}"
        res = self.connector.patchData(
            self.endpoint + path, data=obj, headers=self.header
        )
        return res

    def enableSchedule(self, scheduleId: str = None) -> dict:
        """
        Enable a scheduled query.
        Arguments:
            scheduleId : REQUIRED : id of the schedule.
        """
        if scheduleId is None:
            raise Exception("Missing scheduleId")
        if self.loggingEnabled:
            self.logger.debug(f"Starting enableSchedule")
        obj = {"body": [{"op": "replace", "path": "/state", "value": "enable"}]}
        path = f"/schedules/{scheduleId}"
        res = self.connector.patchData(
            self.endpoint + path, data=obj, headers=self.header
        )
        return res

    def updateSchedule(self, scheduleId: str = None, update_obj: list = None) -> dict:
        """
        Update the schedule query with the object pass.
        Arguments:
            scheduleId : REQUIRED : id of the schedule.
            update_obj : REQUIRED : List of patch operations
        """
        if scheduleId is None:
            raise Exception("Missing scheduleId")
        if update_obj is None:
            raise Exception("Missing update_obj to generate the operation.")
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateSchedules")
        path = f"/schedules/{scheduleId}"
        res = self.connector.patchData(
            self.endpoint + path, data=update_obj, headers=self.header
        )
        return res

    def getTemplates(self, n_results: int = 1000, **kwargs) -> dict:
        """
        Retrieve the list of template for this instance.
        Arguments:
            n_results : OPTIONAL : number of total results to retrieve. To get them all, use "inf".
        possible kwargs:
            limit : Hint on number of records to fetch per page. default (50)
            orderby : Field to order results by. Supported fields: created, updated. Prepend property name with + for ASC,- for DESC order. Default is -created.
            start : Start value of property specified using orderby.
            property : Comma-separated filters.List of properties that allow filtering:
                    name
                    userId
                    lastUpdatedBy
                operations:
                    ‘~’ (contains). This operator can only be used on the name property.
                    ‘==’ (equal to). This operator can be used on both the userId and the lastUpdatedBy properties.
        more details here : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Query-Templates/get_query_templates
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getTemplates")
        path = "/query-templates"
        n_results = float(n_results)  # changing inf to float inf
        params = {**kwargs}
        params["limit"] = kwargs.get("limit", 50)
        res = self.connector.getData(
            self.endpoint + path, headers=self.header, params=params
        )
        data = res["templates"]
        nextPage = res["_links"].get("next", {}).get("href", "")
        while nextPage != "":
            hrefParams = nextPage.split("?")[1]
            orderBy = re.search("orderby=(.+?)(&|$)", hrefParams)
            start = re.search("start=(.+?)(&|$)", hrefParams)
            params["start"] = start.group(1)
            params["orderby"] = orderBy.group(1)
            res = self.connector.getData(self.endpoint + path, params=params)
            data += res.get("templates", [])
            nextPage = res.get("_links", {}).get("next", {}).get("href", "")
            if len(data) > float(n_results):
                nextPage = ""
        return data

    def getTemplate(self, templateId: str = None) -> dict:
        """
        Retrieve a specific template ID.
        Arguments:
            templateId : REQUIRED : template ID to be retrieved.
        """
        if templateId is None:
            raise ValueError("Require a template ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getTemplate")
        path = f"/query-templates/{templateId}"
        res = self.connector.getData(self.endpoint + path)
        return res

    def deleteTemplate(self, templateId: str = None) -> dict:
        """
        Delete a template based on its ID.
        Arguments:
            templateId : REQUIRED : template ID to be deleted.
        """
        if templateId is None:
            raise ValueError("Require a template ID")
        if self.loggingEnabled:
            self.logger.debug(f"Starting deleteTemplate")
        path = f"/query-templates/{templateId}"
        res = self.connector.deleteData(self.endpoint + path)
        return res

    def createQueryTemplate(self, queryData: dict = None) -> dict:
        """
        Create a query template based on the dictionary passed.
        Arguments:
            queryData : REQUIED : An object that contains "sql", "queryParameter" and "name" keys.
        more info : https://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Query-Templates/create_query_template
        """
        path = "/query-templates"
        if self.loggingEnabled:
            self.logger.debug(f"Starting createTemplate")
        if isinstance(queryData, dict):
            if (
                "sql" not in queryData.keys()
                or "queryParameters" not in queryData.keys()
                or "name" not in queryData.keys()
            ):
                raise KeyError(
                    "Minimum key value are not respected.\nPlease see here for more info :\nhttps://www.adobe.io/apis/experienceplatform/home/api-reference.html#/Query-Templates/create_query_template "
                )
        else:
            raise Exception("expected a dictionary for queryData")
        res = self.connector.postData(
            self.endpoint + path, headers=self.header, data=queryData
        )
        return res


class InteractiveQuery:
    """
    Provide the instance connected to PostgreSQL database and you can return the result directly in your notebook.
    This class requires that you have used connection method in the QueryService.
    The object returned by the connection method should be used when creating this object.

    """

    config_object = {}
    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(self, conn_object: dict = None, loggingObject: dict = None):
        if conn_object is None:
            raise AttributeError(
                "You are missing the conn_object. Use the QueryService to retrieve the object."
            )
        self.dbname = conn_object["dbName"]
        self.host = conn_object["host"]
        self.port = conn_object["port"]
        self.user = conn_object["username"]
        self.passwd = conn_object["token"]
        self.websocketHost = conn_object["websocketHost"]
        self.config_object = {
            "dbname": self.dbname,
            "host": self.host,
            "user": self.user,
            "passwd": self.passwd,
            "port": self.port,
        }
        self.db = DB(**self.config_object)
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)

    def query(
        self, sql: str = None, output: str = "dataframe"
    ) -> Union[pd.DataFrame, object]:
        """
        Query the database and return different type of data, depending the format parameters.
        Requests are limited to return 50 K rows
        Arguments:
            sql : REQUIRED : the SQL request you want to realize.
            output : OPTIONAL : the format you would like to be returned.
            Possible format:
                "raw" : return the instance of the query object.
                "dataframe" : return a dataframe with the data. (default)
        """
        if sql is None:
            raise Exception("Required a SQL query")
        if self.loggingEnabled:
            self.logger.debug(f"Starting query:\n {sql}")
        query = self.db.query(sql)
        if output == "raw":
            return query
        elif output == "dataframe":
            data = query.getresult()
            columns = query.listfields()
            df = pd.DataFrame(data, columns=columns)
            return df
        else:
            raise KeyError("You didn't specify a correct value.")

    def transformToDataFrame(self, query: object = None) -> pd.DataFrame:
        """
        This will return you a dataFrame
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting transformToDataFrame")
        data = query.getresult()
        columns = query.listfields()
        df = pd.DataFrame(data, columns=columns)
        return df

    def queryIdentity(
        self,
        identityId: str = None,
        fields: list = None,
        tableName: str = None,
        output: str = "dataframe",
        fieldId: str = "ECID",
        limit: str = None,
        save: bool = False,
        verbose: bool = False,
    ) -> Union[pd.DataFrame, object]:
        """
        Return the elements that you have passed in field list and return the output selected.
        Arguments:
            identityId : REQUIRED : The ID you want to retrieve
            fields : REQUIRED : a list of fields you want to return for that ID in your table.
                example : ['person.name']
            tableName : REQUIRED : The dataset table name to use
            output : OPTIONAL : the format you would like to be returned.
            Possible format:
                "raw" : return the instance of the query object.
                "dataframe" : return a dataframe with the data. (default)
            fieldId : OPTIONAL : If you want your selection to be based on another field than ECID in IdentityMap.
            limit : OPTIONAL : If you wish to set a LIMIT on number of row returned.
            save : OPTIONAL : will save a csv file
            verbose : OPTIONAL : will display some comment
        """
        if identityId is None:
            raise ValueError("Require an identity value")
        if type(fields) != list:
            raise ValueError("Require a list of fields to be returned")
        if tableName is None:
            raise ValueError("Require a dataset table name")
        if fieldId == "ECID":
            condition = f"WHERE identityMap['ECID'][0].id = '{identityId}'"
        elif fieldId != "ECID" and fieldId is not None:
            condition = f"WHERE {fieldId} = '{identityId}'"
        else:
            condition = ""
        if limit is None:
            limit = ""
        else:
            limit = f"LIMIT {limit}"
        sql = f"SELECT {','.join(fields)} FROM {tableName} {condition} {limit}"
        if verbose:
            print(sql)
        if self.loggingEnabled:
            self.logger.debug(f"Starting queryIdentity:\n {sql}")
        res = self.query(sql=sql, output=output)
        if verbose:
            print(f"Data is returned in the {output} format")
        if save:
            if isinstance(res, pd.DataFrame):
                res.to_csv(f"{identityId}.csv", index=False)
            else:
                data = res.getresult()
                columns = res.listfields()
                df = pd.DataFrame(data, columns=columns)
                df.to_csv(f"{identityId}.csv", index=False)
        return res



import aepp
from aepp import connector
from copy import deepcopy
import logging


class Identity:
    """
    Class to manage and retrieve Identity information.
    #!acpdr/swagger-specs/id-service-api.yaml
    This is based on the following API reference : https://www.adobe.io/apis/experienceplatform/home/api-reference.html
    """

    ## logging capability
    loggingEnabled = False
    logger = None

    def __init__(
        self,
        region: str = "nld2",
        config: dict = aepp.config.config_object,
        header=aepp.config.header,
        loggingObject: dict = None,
        **kwargs,
    ):
        """
        Require a region.
        By default, the NLD2 will be selected. (other choice : va7)
        Additional kwargs will update the header.
        more info : https://docs.adobe.com/content/help/en/experience-platform/identity/api/getting-started.html
        Arguments:
            region : REQUIRED : either nld2 or va7
            loggingObject : OPTIONAL : logging object to log messages.
            config : OPTIONAL : config object in the config module. (DO NOT MODIFY)
            header : OPTIONAL : header object  in the config module. (DO NOT MODIFY)
        """
        if loggingObject is not None and sorted(
            ["level", "stream", "format", "filename", "file"]
        ) == sorted(list(loggingObject.keys())):
            self.loggingEnabled = True
            self.logger = logging.getLogger(f"{__name__}")
            self.logger.setLevel(loggingObject["level"])
            formatter = logging.Formatter(loggingObject["format"])
            if loggingObject["file"]:
                fileHandler = logging.FileHandler(loggingObject["filename"])
                fileHandler.setFormatter(formatter)
                self.logger.addHandler(fileHandler)
            if loggingObject["stream"]:
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(formatter)
                self.logger.addHandler(streamHandler)
        self.connector = connector.AdobeRequest(
            config_object=config,
            header=header,
            loggingEnabled=self.loggingEnabled,
            logger=self.logger,
        )
        self.header = self.connector.header
        self.header.update(**kwargs)
        self.sandbox = self.connector.config["sandbox"]
        self.endpoint = (
            f"https://platform-{region}.adobe.io" + aepp.config.endpoints["identity"]
        )

    def getIdentity(
        self, id_str: str = None, nsid: str = None, namespace: str = None
    ) -> dict:
        """
        Given the namespace and an ID in that namespace, returns XID string.
        Arguments:
            id_str : REQUIRED : Id in given namespace (ECID value)
            nsid : REQUIRED : namespace id. (e.g. 411)
            namespace : OPTIONAL : namespace code (e.g. adcloud)
        """
        if id_str is None or nsid is None:
            raise Exception(
                "Expecting that id_str and namespace arguments to be filled."
            )
        if self.loggingEnabled:
            self.logger.debug(f"Starting getIdentity")
        params = {"id": id_str, "nsid": nsid}
        if namespace is not None:
            params["namespace"] = namespace
        path = "/identity/identity"
        self.header["Accept"] = "application/vnd.adobe.identity+json;version=1.2"
        self.header["x-uis-cst-ctx"] = "stub"
        res = self.connector.getData(
            self.endpoint + path, headers=self.header, params=params
        )
        del self.header["x-uis-cst-ctx"]
        self.header["Accept"] = "application/json"
        return res

    def getIdentities(self, only_custom: bool = False, save: bool = False) -> list:
        """
        Get the list of all identity namespaces available in the organization.
        Arguments:
            only_custom : OPTIONAL : if set to True, return only customer made identities (default False)
            save : OPTIONAL : if set to True, save the result in its respective folder (default False)
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getIdentities")
        path = "/idnamespace/identities"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        if only_custom:
            res = [identity for identity in res if identity["custom"] == True]
        if save:
            aepp.saveFile(
                module="identity", file=res, filename="identities", type_file="json"
            )
        return res

    def getIdentityDetail(self, id_str: str = None, save: bool = False) -> dict:
        """
        List details of a specific identity namespace by its ID.
        Arguments:
            id_str : REQUIRED : identity of the "id" field.
            save : OPTIONAL : if set to True, save the result in a file, in its respective folder (default False)
        """
        if id_str is None:
            raise Exception("Expected an id for the Identity")
        if self.loggingEnabled:
            self.logger.debug(f"Starting getIdentityDetail")
        path = f"/idnamespace/identities/{id_str}"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        if save:
            filename = f"identity_{res['code']}"
            aepp.saveFile(
                module="identity", file=res, filename=filename, type_file="json"
            )
        return res

    def createIdentity(
        self,
        name: str = None,
        code: str = None,
        idType: str = None,
        description: str = None,
        dict_identity: dict = None,
    ) -> dict:
        """
        List details of a specific identity namespace by its ID.
        Arguments:
            name : REQUIRED : Display name of the identity
            code : REQUIRED : Identity Symbol for user interface.
            idType : REQUIRED : one of those : COOKIE, CROSS_DEVICE, DEVICE, EMAIL, MOBILE, NON_PEOPLE or PHONE.
            description : OPTIONAL : description for this identity
            dict_identity : OPTIONAL : you can use this to directly pass the dictionary.
        """
        creation_dict = {}
        if name is None or code is None or idType is None:
            raise Exception(
                "Expecting that name, code and idType to be filled with value"
            )
        creation_dict["name"] = name
        creation_dict["code"] = code
        creation_dict["idType"] = idType
        if description is not None:
            creation_dict["description"] = description
        if " " in code:
            raise TypeError("code can only contain one word with letter and numbers")
        if idType not in [
            "COOKIE",
            "CROSS_DEVICE",
            "DEVICE",
            "EMAIL",
            "MOBILE",
            "NON_PEOPLE",
            "PHONE",
        ]:
            raise TypeError(
                "idType could only be one of those : COOKIE, CROSS_DEVICE, DEVICE, EMAIL, MOBILE, NON_PEOPLE, PHONE"
            )
        if self.loggingEnabled:
            self.logger.debug(f"Starting createIdentity")
        if dict_identity is not None:
            creation_dict = dict_identity
        path = "/idnamespace/identities"
        res = self.connector.postData(
            self.endpoint + path, headers=self.header, data=creation_dict
        )
        return res

    def updateIdentity(
        self,
        id_str: str = None,
        name: str = None,
        code: str = None,
        idType: str = None,
        description: str = None,
    ) -> dict:
        """
        Update identity based on its ID.
        Arguments:
            id_str: REQUIRED : ID of the identity namespace to update.
            name : REQUIRED : Display name of the identity
            code : REQUIRED : Identity Symbol for user interface.
            idType : REQUIRED : one of those : COOKIE, CROSS_DEVICE, DEVICE, EMAIL, MOBILE, NON_PEOPLE or PHONE.
            description : OPTIONAL : description for this identity
        """
        if id_str is None:
            raise Exception("Require an id")
        if name is None or code is None or idType is None:
            raise Exception(
                "Expecting that name, code and idType to be filled with value"
            )
        if idType not in [
            "COOKIE",
            "CROSS_DEVICE",
            "DEVICE",
            "EMAIL",
            "MOBILE",
            "NON_PEOPLE",
            "PHONE",
        ]:
            raise TypeError(
                "idType could only be one of those : COOKIE, CROSS_DEVICE, DEVICE, EMAIL, MOBILE, NON_PEOPLE, PHONE"
            )
        if self.loggingEnabled:
            self.logger.debug(f"Starting updateIdentity")
        path = f"/idnamespace/identities/{id_str}"
        data = {
            "name": name,
            "code": code,
            "idType": idType,
            "description": description,
        }
        res = self.connector.putData(
            self.endpoint + path, headers=self.header, data=data
        )
        return res

    def getIdentitiesIMS(self, imsOrg: str = None) -> list:
        """
        Returns all identities from the IMS Org itself.
        Only shared ones if IMS Org doesn't match the IMS Org sent in the header.
        Arguments:
            imsOrg : OPTIONAL : the IMS org. If not set, takes the current one automatically.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getIdentitiesIMS")
        ims_org = imsOrg or self.connector.config["org_id"]
        path = f"/idnamespace/orgs/{ims_org}/identities"
        res = self.connector.getData(self.endpoint + path, headers=self.header)
        return res

    def getClustersMembers(
        self,
        xid: str = None,
        nsid: str = "411",
        namespace: str = "adcloud",
        id_value: str = None,
        graphType: str = "private",
    ) -> dict:
        """
        Given an XID return all XIDs, in the same or other namespaces, that are linked to it by the device graph type.
        The related XIDs are considered to be part of the same cluster.
        It is required to pass either xid or (namespace/nsid & id) pair to get cluster members.
        Arguments:
            xid : REQUIRED : Identity string returns by the getIdentity method.
            nsid : OPTIONAL : namespace id (default : 411)
            namespace : OPTIONAL : namespace code. (default : adcloud)
            id_value : OPTIONAL : ID of the customer in given namespace.
            graphType : OPTIONAL : Graph type (output type) you want to get the cluster from. (default private)
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getClustersMembers")
        temp_header = deepcopy(self.header)
        temp_header["Accept"] = "application/vnd.adobe.identity+json;version=1.2"
        temp_header["x-uis-cst-ctx"] = "stub"
        path = "/identity/cluster/members"
        params = {}
        if xid is not None:
            params["xid"] = xid
            params["graph-type"] = graphType
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=temp_header
            )
            return res
        elif xid is None and id_value is not None:
            params["nsid"] = nsid
            params["namespace"] = namespace
            params["id"] = id_value
            params["graph-type"] = graphType
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=temp_header
            )
            return res

    def postClustersMembers(
        self, xids: list = None, version: float = 1.0, graphType: str = "private"
    ) -> dict:
        """
        Given set of identities, returns all linked identities in cluster corresponding to each identity.
        Arguments:
            xids : REQUIRED : list of identity as returned by getIdentity method.
            version : OPTIONAL : Version of the clusterMembers (default 1.0)
            graphType : OPTIONAL : Graph type (output type) you want to get the cluster from. (default private)
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting postClustersMembers")
        temp_header = deepcopy(self.header)
        temp_header["Accept"] = "application/vnd.adobe.identity+json;version=1.2"
        temp_header["x-uis-cst-ctx"] = "stub"
        path = "/identity/cluster/members"
        if type(xids) != list:
            raise TypeError("xids must be of type list")
        list_body = [
            {"xid": [{"xid": xid}], "graph-type": graphType, "version": version}
            for xid in xids
        ]
        res = self.connector.postData(
            self.endpoint + path, data=list_body, headers=temp_header
        )
        return res

    def getClusterHistory(
        self,
        xid: str = None,
        nsid: int = 411,
        namespace: str = "adcloud",
        id_value: str = None,
        graphType: str = "private",
    ) -> dict:
        """
        Given an XID, return all cluster associations with that XID.
        It is required to pass either xid or (namespace/nsid & id) pair to get cluster history.
        Arguments:
            xid : REQUIRED : Identity string returns by the getIdentity method.
            nsid : OPTIONAL : namespace id (default : 411)
            namespace : OPTIONAL : namespace code. (default : adcloud)
            id_value : OPTIONAL : ID of the customer in given namespace.
            graphType : OPTIONAL : Graph type (output type) you want to get the cluster from. (default private)
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getClusterHistory")
        temp_header = deepcopy(self.header)
        temp_header["Accept"] = "application/vnd.adobe.identity+json;version=1.2"
        temp_header["x-uis-cst-ctx"] = "stub"
        path = "/identity/cluster/history"
        params = {}
        if xid is not None:
            params["xid"] = xid
            params["graph-type"] = graphType
            res = aepp._getData(
                self.endpoint + path, params=params, headers=temp_header
            )
            return res
        elif xid is None and id_value is not None:
            params["nsid"] = nsid
            params["namespace"] = namespace
            params["id"] = id_value
            params["graph-type"] = graphType
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=temp_header
            )
            return res

    def getIdentityMapping(
        self,
        xid: str = None,
        targetNs: int = None,
        nsid: int = 411,
        namespace: str = "adcloud",
        id_value: str = None,
        graphType: str = "private",
    ) -> dict:
        """
        Given an XID, returns all XID mappings in the requested namespace (targetNs).
        It is required to pass either xid or (namespace/nsid & id) pair to get mappings in required namespace.
        Arguments:
            xid : REQUIRED : Identity string returns by the getIdentity method.
            nsid : OPTIONAL : namespace id (default : 411)
            namespace : OPTIONAL : namespace code. (default : adcloud)
            id_value : OPTIONAL : ID of the customer in given namespace.
            graphType : OPTIONAL : Graph type (output type) you want to get the cluster from. (default private)
            targetNs : OPTIONAL : The namespace you want to get the mappings from.
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting getIdentityMapping")
        temp_header = deepcopy(self.header)
        temp_header["Accept"] = "application/vnd.adobe.identity+json;version=1.2"
        temp_header["x-uis-cst-ctx"] = "stub"
        path = "/identity/mapping"
        params = {}
        if xid is not None:
            params["xid"] = xid
            params["graph-type"] = graphType
            params["targetNs"] = targetNs
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=temp_header
            )
            return res
        elif xid is None and id_value is not None:
            params["nsid"] = nsid
            params["namespace"] = namespace
            params["id"] = id_value
            params["targetNs"] = targetNs
            params["graph-type"] = graphType
            res = self.connector.getData(
                self.endpoint + path, params=params, headers=temp_header
            )
            return res

    def postIdentityMapping(
        self, xids: list = None, targetNs: int = 411, version: float = 1.0
    ) -> dict:
        """
        Given an identity, returns all identity mappings in requested namespace (target namespace).
        Arguments:
            xids : REQUIRED : List of identities
            targetNs : REQUIRED : Target Namespace (default 411)
            version : OPTIONAL : version of the mapping
        """
        if self.loggingEnabled:
            self.logger.debug(f"Starting postIdentityMapping")
        temp_header = deepcopy(self.header)
        temp_header["Accept"] = "application/vnd.adobe.identity+json;version=1.2"
        temp_header["x-uis-cst-ctx"] = "stub"
        path = "/identity/mapping"
        if type(xids) != list:
            raise TypeError("xids must be of type list")
        list_body = [
            {"xid": [{"xid": xid}], "version": version, "targetNs": targetNs}
            for xid in xids
        ]
        res = self.connector.postData(
            self.endpoint + path, data=list_body, headers=temp_header
        )
        return res
